/*
 * $Id: qg_util.c,v 1.25 2003/10/07 01:59:27 ldemailly Exp $
 *
 * Copyright (c) 2003, Qualys, Inc.  All rights reserved.
 *
 * See the file LICENSE for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "libqgInt.h"

/**
 * qg_recv is needed by curl to receive something 
 * returns -1 on error and the size of the received data on success
 */
size_t qg_recv(void *buffer, size_t size, size_t nmemb, void *userp)
{
    size_t oldsize;
    qgresponse_t *qgh = NULL;

    if (!buffer || !userp) {
        qg_error(0, LIBQG_RECV, LIBQG_ERR_MISPAR, "buffer / userp");
        return (-1);
    }

    qgh = (qgresponse_t *) userp;

    // Expand char array
    oldsize = qgh->size;
    qgh->size += nmemb;
    // +1 so there is room for adding a 0 at the end
    if ((qgh->data = (char *) realloc(qgh->data, qgh->size + 1)) == NULL) {
        qg_error(0, LIBQG_RECV, LIBQG_ERR_REALLOC);
        return (-1);
    }
    // Append data
    memcpy((qgh->data + oldsize), buffer, nmemb);
    // Null terminate (if result will be treated as string)
    qgh->data[qgh->size] = 0;

    return (nmemb);
}


/**
 * qg_fetch_raw connects to the specified url
 * returns NULL on error, raw data otherwise
 */
qgresponse_t *qg_fetch_raw(qghandle_t * qgh, char *url)
{
    char errbuf[CURL_ERROR_SIZE];
    qgresponse_t *qgr = NULL;

    if (!qgh || !url) {
        qg_error(0, LIBQG_FETCH, LIBQG_ERR_MISPAR, "qgh / url");
        return (NULL);
    }

    // Set url & fetch
    curl_easy_setopt(qgh->curl, CURLOPT_URL, url);

    // Allocate space for response
    if ((qgr = (qgresponse_t *) malloc(sizeof(qgresponse_t))) == NULL) {
        qg_error(0, LIBQG_FETCH, LIBQG_ERR_MALLOC);
        return (NULL);
    }

    qgr->data = NULL;
    qgr->size = 0;
    qgr->xml = NULL;

    // Setup handler to recieve data
    curl_easy_setopt(qgh->curl, CURLOPT_WRITEFUNCTION, qg_recv);
    curl_easy_setopt(qgh->curl, CURLOPT_FILE, qgr);

    curl_easy_setopt(qgh->curl, CURLOPT_ERRORBUFFER, errbuf);

    libqg_message(0, LIBQG_FETCH_URL, url);

    // Get page
    qgr->res = curl_easy_perform(qgh->curl);

    if (qgr->res != 0) {
        // Detect bad certificates :
        qg_error(0, LIBQG_FETCH, LIBQG_ERR_CURL, qgr->res, url, errbuf);
        if (qgr->res == CURLE_SSL_CACERT || 
            qgr->res == CURLE_SSL_PEER_CERTIFICATE || 
            qgr->res == CURLE_SSL_CONNECT_ERROR) {
            qg_error(1, LIBQG_FETCH, LIBQG_ERR_SSL);
        }
        return NULL;
    }

    curl_easy_getinfo(qgh->curl, CURLINFO_HTTP_CODE, &(qgr->httpCode));
    if (qgr->httpCode >= HTTP_ERRORS_START) {
        libqg_message(FALSE, "Error return code %d for %s", qgr->httpCode, url);
    }

    return (qgr);
}

/**
 * qg_fetch calls qg_fetch_raw and parses the XML-data, if XML-data
 * was received
 * returns NULL on error, xml data (guaranteed not empty) otherwise
 */
qgresponse_t *qg_fetch(qghandle_t * qgh, char *url)
{
    int error_bool = 0;
    int failed_bool = 0;
    char *failerrmsg = NULL; 
    xmlXPathContextPtr context = NULL;
    xmlXPathObjectPtr errorObj = NULL;
    xmlXPathObjectPtr failedObj = NULL;
    xmlNodePtr errorNode = NULL;
    qgresponse_t *qgr = qg_fetch_raw(qgh, url);
    
    if (qgr == NULL) {
        // Error message/code already set by fetch_raw
        return NULL;
    }
    if (qgr->httpCode >= HTTP_ERRORS_START) {
        if (qgr->httpCode == HTTP_AUTH_FAIL) {
            qg_error(1, LIBQG_VERI_LOG, "You have specified a wrong username or password.\n    Please check your input and try again.");
        }
        qg_free_response(qgr);
        return NULL;
    }

    // Build XML if possible
    if ((qgr->size <= 0) || (qgr->data == NULL)) {
        qg_error(FALSE, LIBQG_FETCH, "Empty document returned by QG (for %s)", url);
        qg_free_response(qgr);
        return NULL;
    }

    qgr->xml = xmlParseMemory(qgr->data, qgr->size);

    if (!qgr->xml) {
        qg_error(FALSE, LIBQG_FETCH, "Invalid XML document");
        qg_free_response(qgr);
        return NULL;
    }

    context = xmlXPathNewContext(qgr->xml);
    errorObj = xmlXPathEvalExpression("//ERROR", context);
    failedObj = xmlXPathEvalExpression("//RETURN[@status=\"FAILED\"]", context);
    
    if (!xmlXPathNodeSetIsEmpty(errorObj->nodesetval)) {
        error_bool = 1;
        errorNode = errorObj->nodesetval->nodeTab[0]->children;
    }

    if (!xmlXPathNodeSetIsEmpty(failedObj->nodesetval)) {
        failed_bool = 1;
        errorNode = failedObj->nodesetval->nodeTab[0]->children;
    }

    if (error_bool || failed_bool) {
        failerrmsg = xmlNodeListGetString(qgr->xml, errorNode, 1);
        qg_error(0, LIBQG_FETCH, 
                 "%s contained an error.\n    Error-message:%s",
                 url, failerrmsg);
        xmlFree(failerrmsg);
        qg_free_response(qgr);
        qgr = NULL;
    }
            
    xmlXPathFreeObject(errorObj);
    xmlXPathFreeObject(failedObj);
    xmlXPathFreeContext(context);
    return (qgr);
}

/**
 * qg_free_response frees the data-structure returned by qg_fetch_raw
 */
void qg_free_response(qgresponse_t * qgr)
{
    if (!qgr) {
        return;
    }

    // Free XML
    if (qgr->xml) {
        xmlFreeDoc(qgr->xml);
    }

    // Free Data Ptr
    if (qgr->size && qgr->data) {
        free(qgr->data);
    }

    free(qgr);
    return;
}

/**
 * qg_strappend appends the string specified by 'addstr' to 'base'
 * returns NULL on error
 */
char *qg_strappend(char *base, char *addstr)
{
    int oldsize = 0;
    int newsize = 0;
    char *newbuf = NULL;

    if (!base || !addstr) {
        qg_error(0, LIBQG_STRAP, LIBQG_ERR_MISPAR, "base / addstr");
        return (NULL);
    }

    // Expand char array
    oldsize = strlen(base);
    newsize = (oldsize + strlen(addstr) + 1);   // We add one for zero term
    if ((newbuf = (char *) realloc(base, newsize)) == NULL) {
        qg_error(0, LIBQG_STRAP, LIBQG_ERR_REALLOC);
        return (NULL);
    }
    // Append data
    strcat(newbuf, addstr);
    return (newbuf);
}

/**
 * qg_error saves an error in qg_err_message, and if LIBQG_DEBUG is
 * enabled prints out the name of the function where the error has
 * occurred.
 */
void qg_error(int append, char *usedfunc, const char *error,...)
{
    char temp_buf[STR_BUF_LEN];
    va_list ap;
    va_start(ap, error);
    vsnprintf(temp_buf, STR_BUF_LEN, error, ap);
    if (LIBQG_DEBUG) {
        printf("[E] %s: %s\n", usedfunc, temp_buf);
    }
    if (qg_err_message != NULL) {
        if (append) {
            qg_err_message = qg_strappend(qg_err_message, "\n    ");
            qg_err_message = qg_strappend(qg_err_message, temp_buf);
        } else {
            free(qg_err_message);
            qg_err_message = strdup(temp_buf);
        }
    } else {
        qg_err_message = strdup(temp_buf);
    }
}

/**
 * strqg_error is for getting the errormessage outside of libqg
 * it returns the errormessage
 */
char *strqg_error(char *message)
{
    char temp_buf[STR_BUF_LEN];

    if(qg_err_message == NULL) {
        return message;
    } else {
        snprintf(temp_buf, STR_BUF_LEN, "%s\n    libqg: %s\n", 
                 message, qg_err_message);
        return strdup(temp_buf);
    }
}

void libqg_message(int type, const char *message,...)
{
    char temp_buf[STR_BUF_LEN];
    char temp[STR_BUF_LEN];
    va_list ap;

    if (!qg_debug) {
        return;
    }

    va_start(ap, message);
    vsnprintf(temp_buf, STR_BUF_LEN, message, ap);

    switch (type) {
        case DBG:
            snprintf(temp, STR_BUF_LEN, "[DQG] %s\n", temp_buf);
            break;
        default:
            return;
    }

    printf(temp);
}
