/*
 * $Id: qg_xmlutil.c,v 1.31 2003/10/06 23:33:55 ldemailly Exp $
 *
 * Copyright (c) 2003, Qualys, Inc.  All rights reserved.
 *
 * See the file LICENSE for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "libqgInt.h"

/**
 * qg_xml_find_os searches for the tags
 * <SERVICES>, <CAT value="Information">, <SERVICE value="os_exact">
 * returns the content of os_exact if the XML-document contains
 * this field, and NULL if it is not found 
 */
char *qg_xml_find_os(xmlNodePtr iproot)
{
    xmlNodePtr cur = NULL;

    if (!iproot) {
        return (NULL);
    }

    if ((cur = qg_xml_find_child(iproot, XML_SERV)) == NULL) {
        return (NULL);
    }

    if ((cur = qg_xml_find_prop(cur, XML_VALU, "Information")) == NULL) {
        return (NULL);
    }

    if ((cur = qg_xml_find_prop(cur, XML_VALU, "os_exact")) == NULL) {
        return (NULL);
    }

    if ((cur = qg_xml_find_child(cur, XML_RESU)) == NULL
        || cur->children == NULL) {
        return (NULL);
    }

    return (cur->children->content);
}

/**
 * qg_xml_get_report_ref returns the first reference-code
 * of a XML-code
 * returns NULL on error, "NOT_FOUND" if no reference-code
 * was found and the code on success
 */
char *qg_xml_get_report_ref(xmlDocPtr doc)
{
    xmlNodePtr cur = NULL;
    char *ref = NULL;

    if (!doc) {
        qg_error(0, LIBQG_XMLU_FETREP, LIBQG_ERR_MISPAR, "doc");
        return (NULL);
    }

    if ((cur = xmlDocGetRootElement(doc)) == NULL) {
        qg_error(0, LIBQG_XMLU_FETREP, "no XML-root element");
        return (NULL);
    }

    for (cur = cur->children; cur != NULL; cur = cur->next) {
        if ((ref = xmlGetProp(cur, "ref")) != NULL) {
            return ref;
        }
    }
    return LIBQG_NOT_FOUND;
}

/**
 * qg_custom_report creates drops all for correlation unnecessary information
 * from the QualysGuard-report and creates its own structure for easy getting
 * information about the IP 
 */
xmlDocPtr qg_custom_report(xmlDocPtr report, char *ip, char *cachedir)
{
    double vulncover = 0;
    double threatcover = 0;
    double total = 0;
    char *qtifile = strdup(cachedir);
    char vulnstr[SHORT_BUF];
    char threatstr[SHORT_BUF];
    char totalstr[SHORT_BUF];
    custrep_t *crep = calloc(1, sizeof(custrep_t));
    xmlDocPtr custom = NULL;
    xmlDocPtr qtidoc = NULL;
    xmlNodePtr customroot = NULL;
    xmlNodePtr newstat = NULL;
    xmlNodePtr stats = NULL;
    xmlNodePtr vuln = NULL;
    xmlNodePtr info = NULL;
    
    qtifile = qg_strappend(qtifile, "/qtimap.xml");

    // load mapping file
    qtidoc = xmlParseFile(qtifile);
    crep->qgcontext = xmlXPathNewContext(report);
    crep->qidtoids = xmlXPathNewContext(qtidoc);
    crep->ip = ip;
    
    // create QuIDScor-report
    custom = xmlNewDoc("1.0");
    customroot = xmlNewDocNode(custom, NULL, "RESULT", NULL);
    xmlNewProp(customroot, "IP", crep->ip);
    xmlDocSetRootElement(custom, customroot);

    stats = xmlNewChild(customroot, NULL, "STATS", NULL);
    info = xmlNewChild(customroot, NULL, "INFOS", NULL);
    vuln = xmlNewChild(customroot, NULL, "VULNS", NULL);
    
    // parse the QualysGuard-report
    qg_xml_custom_port(crep->qgcontext, info, QG_TCP, crep->ip);
    qg_xml_custom_port(crep->qgcontext, info, QG_UDP, crep->ip);
    qg_xml_custom_procvuln(crep, vuln, QG_VULN);
    qg_xml_custom_procvuln(crep, vuln, QG_THREAT);

    // coverage statistics:
    // coverage of vulns
    if (crep->vulnmap != 0) {
        vulncover = crep->vulnmap / crep->vuln * 100.00;
    }
    newstat = xmlNewChild(stats, NULL, "COVERAGE", NULL);
    xmlNewProp(newstat, "TYPE", "VULNS");
    snprintf(vulnstr, SHORT_BUF, "%3.2f", vulncover);
    xmlNewProp(newstat, "PERCENTAGE", vulnstr);

    //coverage of threats
    if (crep->threatmap != 0) {
        threatcover = crep->threatmap / crep->threat * 100.00;
    }
    newstat = xmlNewChild(stats, NULL, "COVERAGE", NULL);
    xmlNewProp(newstat, "TYPE", "THREATS");
    snprintf(threatstr, SHORT_BUF, "%3.2f", threatcover);
    xmlNewProp(newstat, "PERCENTAGE", threatstr);
    
    // total coverage
    if (crep->threatmap != 0 || crep->vulnmap != 0) {
        total = (crep->threatmap + crep->vulnmap) / (crep->threat + crep->vuln) * 100;
    }
    newstat = xmlNewChild(stats, NULL, "COVERAGE", NULL);
    xmlNewProp(newstat, "TYPE", "TOTAL");
    snprintf(totalstr, SHORT_BUF, "%3.2f", total);
    xmlNewProp(newstat, "PERCENTAGE", totalstr);

    // free custom-report-structure
    xmlXPathFreeContext(crep->qgcontext);
    xmlXPathFreeContext(crep->qidtoids);
    xmlFreeDoc(qtidoc);
    free(crep);
    free(qtifile);
    return custom;
}

/**
 * qg_xml_custom_procvuln searches for all necessary information concerning
 * vulnerabilites and threats
 */
void qg_xml_custom_procvuln(custrep_t *crep, xmlNodePtr report, QG_custrep type) 
{
    xmlNodePtr customvuln = NULL;
    xmlXPathObjectPtr qidsearch = NULL;
    xmlXPathObjectPtr searchptr = NULL;
    char searchpath[STR_BUF_LEN];
    char qidpath [STR_BUF_LEN];
    char *searchstr = NULL;
    char *qid = NULL;
    char *sev = NULL;
    char *port = NULL;
    char *typestr = NULL;
    int count = 0;
    int nslen = 0;
    int i = 0;

    switch (type) {
    case QG_VULN:
        searchstr = XMLPATH_VULN;
        typestr = "VULN";
        break;
    default:
    case QG_THREAT:
        searchstr = XMLPATH_THREAT;
        typestr = "THREAT";
        break;
    }
    
    // search for all Vulns/Threats in the report
    snprintf(searchpath, STR_BUF_LEN, searchstr, crep->ip);
    searchptr = xmlXPathEvalExpression(searchpath, crep->qgcontext);

    if ((nslen = xmlXPathNodeSetGetLength(searchptr->nodesetval)) == 0) {
        xmlXPathFreeObject(searchptr);
        return;
    }

    // walk through all elements of the searchresult
    while (i < nslen) {
        xmlNodePtr next = searchptr->nodesetval->nodeTab[i++];

        // get ID of the Vuln/Threat
        qid = xmlGetProp(next, "number");

        // create entry for new report
        customvuln = xmlNewChild(report, NULL, "FOUND", NULL);
        xmlNewProp(customvuln, "ID", qid);
        xmlNewProp(customvuln, "TYPE", typestr);

        // if severity is specified add to custom report
        if ((sev = xmlGetProp(next, "severity")) != NULL) {
            xmlNewProp(customvuln, "SEV", sev);
            xmlFree(sev);
            sev = NULL;
        }

        // if port is specified add to custom report
        if ((port = xmlGetProp(next->parent, "port")) != NULL) {
            xmlNewProp(customvuln, "PORT", port);
            xmlFree(port);
            port = NULL;
        }
        
        // search in mapping-table for mapped ID
        snprintf(qidpath, STR_BUF_LEN,"/QIDTOIDS/IN[@ID=\"%s\"]",qid);
        qidsearch = xmlXPathEvalExpression(qidpath, crep->qidtoids);

        // set flag if mapping exists
        if (!xmlXPathNodeSetIsEmpty(qidsearch->nodesetval)) {
            xmlNewProp(customvuln, "MAPPED", "true");
            count++;
        } else {
            xmlNewProp(customvuln, "MAPPED", "false");
        }

        if (qidsearch) {
            xmlXPathFreeObject(qidsearch);
            qidsearch = NULL;
        }

        // free ID
        xmlFree(qid);
        qid = NULL;
    }
    xmlXPathFreeObject(searchptr);

    // save statistical data in custom-report structure
    switch (type) {
    case QG_VULN:
        crep->vuln = (double)nslen;
        crep->vulnmap = (double)count;
        break;
    default:
    case QG_THREAT:
        crep->threat = (double)nslen;
        crep->threatmap = (double)count;
        break;
    }
} 

/**
 * qg_xml_custom_port searches for the strings containing port information
 * and saves the information in XML-tags
 */
void qg_xml_custom_port(xmlXPathContextPtr qgcontext, xmlNodePtr parent, QG_custrep protocol, char *ip)
{
    xmlNodePtr port = NULL;
    xmlXPathObjectPtr searchptr = NULL;
    char searchpath[STR_BUF_LEN];
    char *searchstr = NULL;
    char *linebuf = NULL;
    char *linerun = NULL;
    char *linestart = NULL;
    char *protstr = NULL;
    int line = 0;
    int col = 0;
    int str = 0;

    // set protocol type for search
    switch (protocol) {
    case QG_TCP:
        searchstr = XMLPATH_TCP;
        protstr = "TCP";
        break;
    default:
    case QG_UDP:
        searchstr = XMLPATH_UDP;
        protstr = "UDP";
        break;
    }

    // search for port-information
    snprintf(searchpath, STR_BUF_LEN, searchstr, ip);
    searchptr = xmlXPathEvalExpression(searchpath, qgcontext);

    // no result
    if (xmlXPathNodeSetIsEmpty(searchptr->nodesetval)) {
        xmlXPathFreeObject(searchptr);
        return;
    }
    
    // no contents
    if (searchptr->nodesetval->nodeTab[0]->children->content == NULL) {
        xmlXPathFreeObject(searchptr);
        return;
    }

    linebuf = strdup(searchptr->nodesetval->nodeTab[0]->children->content);
    linerun = linebuf;

    // walk through whole string
    while (*linerun != '\0') {
        // set start of string
        if (str == 0) {
            linestart = linerun;
            str = 1;
        }

        // ignore the first line and check only the first four columns
        // use this code also only if the end of a column is reached --> '\t'
        if (line > 0 && col < 4 && *linerun == '\t'){
            // replace tab with \0
            *linerun = 0;

            // just in case the last char was a space
            if (*(linerun - 1) == ' ') {
                *(linerun - 1) = 0;
            }
            
            // add properties to port-element
            if (col == 0 && line != 0) {
                xmlNewProp(port, "VALUE", linestart);
            } else if (col == 3 && line != 0) {
                xmlNewProp(port, "SERVICE", linestart);
                // check for application name of webserver
                if (strstr(linestart, "http") != NULL) {
                    qg_xml_get_webservname(qgcontext, port, ip);
                }
            }
            
            str = 0;
            col++;
        }
        
        // if end of line is reached, create port-element for custom-report
        if(*linerun == '\n') {
            port = xmlNewChild(parent, NULL, "PORT", NULL);
            xmlNewProp(port, "PROTOCOL", protstr);
            
            // just in case next line is the end of the string
            if (*(linerun + 1) != '\0') {
                // set start element to start of first column in next line
                linestart = linerun + 1;
            }
            
            line++;

            // reset string and column-counter
            str = 0;
            col = 0;
        }
        
        linerun++;
    }
    // add last property to last port-element
    xmlNewProp(port, "SERVICE", linestart);

    xmlXPathFreeObject(searchptr);
    free(linebuf);
}


/**
 * qg_xml_get_webservname is a first implementation to add the application name
 * to the QuIDScor-reports
 */
void qg_xml_get_webservname(xmlXPathContextPtr qgcontext, xmlNodePtr port, char *ip)
{
    char *portstr = xmlGetProp(port, "VALUE");
    char searchbuf[STR_BUF_LEN];
    char *linebuf = NULL;
    int offset = 0;
    int tabcounter = 1;
    int tabPosition = 0;
    int reachedLine = FALSE;
    int colReached = FALSE;
    int startServerName = 0;
    xmlXPathObjectPtr searchres = NULL;

    snprintf(searchbuf, STR_BUF_LEN, XMLPATH_WEBSERV, ip, portstr);
    searchres = xmlXPathEvalExpression(searchbuf, qgcontext);

    if (xmlXPathNodeSetIsEmpty(searchres->nodesetval)) {
        xmlXPathFreeObject(searchres);
        xmlFree(portstr);
        return;
    }

    if (searchres->nodesetval->nodeTab[0]->children->content == NULL) {
        xmlXPathFreeObject(searchres);
        xmlFree(portstr);
        return;
    }

    linebuf = strdup(searchres->nodesetval->nodeTab[0]->children->content);
    tabPosition = findInFirstColumn(linebuf,"Server Type");
    do  {
        if (!reachedLine && linebuf[offset] == '\n') {
            reachedLine = TRUE;
        }

        if (colReached && (linebuf[offset] == '\t'||
                           linebuf[offset] == '\0'||
                           linebuf[offset] == '\n')) {
            *(linebuf+offset) = '\0';
            xmlNewProp(port, "APPLICATION", linebuf + startServerName);
            break;
        }
        if ((tabPosition == 0) && reachedLine && !colReached) {
            colReached = TRUE;
            startServerName = offset + 1;
        }
        if (reachedLine && !colReached && linebuf[offset] == '\t') {
            if (tabcounter == tabPosition) {
                colReached = TRUE;
                startServerName = offset + 1;
            }
            tabcounter++;
        }

    } while(linebuf[offset++] != '\0');
    
    

    free(linebuf);
    xmlFree(portstr);
    xmlXPathFreeObject(searchres);
}

/**
 * findInFirstColumn finds a String in a Text located in the first 
 * and returns the Index of the Column where it is located
 */
int findInFirstColumn(char* wholeString,char* searchString) {
    int column = 0;
    int start = 0;
    int end = 0;
    int i = 0;
    int success = FALSE;
    while (!success) {
  	end = start + findColEnd(wholeString + start);
        if (wholeString[end] == '\n' || wholeString[end] == '\0') {
            return -1;
        }

        if (strlen(searchString) == (end - start)) {
            success = TRUE;
            for (i = 0; i < (end - start); i++) {
                if (!(wholeString[start + i] == searchString[i])) {
                    success = FALSE;
                }	
            }
        }

        if (!success){
            start += end + 1;
            column++;
        }
    }
    return column;
    
}



/**
 * findColEnd returns the length of the next word
*/
int findColEnd(char* string) {
    int i = 0;
    int returnValue = -1;
    do {
        switch(*(string + i)) {
            case '\0':
            case '\n':
            case '\t':
                returnValue = i;
                break;
            default:
                i++;
        }
    } while(returnValue == -1);
    return returnValue;
}
