/*
 * $Id: quidscor.c,v 1.71 2003/10/09 18:19:13 ldemailly Exp $
 *
 * Copyright (c) 2003, Qualys, Inc.  All rights reserved.
 *
 * See the file LICENSE for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */
#include "quidscor.h"

#ifndef VARPREFIX
#error "need to define the install varprefix"
#endif


// TODO: limit global variables, move to a struct
// and/or prefix (g_qs_...) to avoid namespace issues.
FILE *good = NULL;
FILE *bad = NULL;
FILE *unverified = NULL;
qghandle_t *handle = NULL;
u_long expiretime = DEF_EXPIRE;
int debug = FALSE;
int daemonmode = FALSE;
char *inputfname = NULL;
char *username = NULL;
char *password = NULL;
char *cachedir = NULL;
char *qgserver = QGSERVER;
char *logdir = NULL;
char *mapfile = NULL;
char *pidfile = NULL;
char *rdir = NULL;
#ifdef QS_DRAGON
int mode = MODE_UNDEFINED;
#else
int mode = MODE_SNORT;
#endif
exec_t *edb = NULL;
char *ipgroup = "All";
int dbRefreshOnly = FALSE;
int dbReadOnly    = FALSE;
pid_t cpid = -1; // Child process ID (split processes mode)
int  inRestart = FALSE;
int  lastsig = 0;  // last signal received (when non zero - stop processing)
/** 
 * display_version displays the version-number, Copyright and license
 * information for QuIDScor
 */
void display_version() {
    printf("*** Qualys IDS Event Correlation engine v%s ***\n", VERSION);
    printf("    Copyright (c) 2003 Qualys Inc. All rights reserved.\n");
    printf("    See the LICENSE file for license terms and a DISCLAIMER OF ALL WARRANTIES.\n");
}

/**
 * Main processing call out.
 */
void processing() {
    // make sure getprotobyname() calls aren't too expensive
    setprotoent(TRUE);

    switch (mode) {
            // Snort Main
        case MODE_SNORT:
            snort_main(inputfname);
            break;

            // Log Main
        case MODE_LOG:
            log_input_main(inputfname);
            break;

#ifdef QS_DRAGON
            // Dragon Main
        case MODE_DRAGON:
            dragon_main();
            break;
#endif
        default:
            printf("Unknown mode %d.\n", mode);
            quid_exit(QSCE_MODE);
            break;
    }

    quid_exit(QSCE_NORMAL);
}


/**
 * When in dual process mode, the parent process enters this
 * loop and sleeps and wakes up only to refresh the VA DB
 */
void va_update_loop() {
    // TODO: close resources unused in this process (and 
    // vice versa in the child.

    // Main loop for parent:
    while (1) {
        // sleep
        sleep(expiretime);
        // wakeup, fetch updates
        quid_message(SYS, "Parent waking up after %ld seconds...", expiretime);
        // Reload data
        if (!load_mappings(handle, mapfile, rdir, mode, cachedir)) {
            quid_message(ERR, strqg_error("Error in mapping reload."));
            quid_exit(QSCE_MAP);
        }
        // tell child to finish current alert and exit:
        quid_message(SYS, "Sending SIGUSR1 to child %d", cpid);
        inRestart = TRUE;
        kill(cpid, SIGUSR1);
        // wait that it is done
        wait(NULL); // TODO: check status
        quid_message(SYS, "Child stopped.");
        inRestart = FALSE;
        lastsig = 0;
        // TODO: refactor with the first time fork() below
        cpid = fork();
        switch (cpid) {
            case -1: // Error
                quid_message(ERR, "Split process could not be started.");
                quid_exit(QSCE_FORK);
            case 0 : // Child
                dbReadOnly = TRUE;
                signal(SIGCHLD, SIG_IGN); // TODO: comment why/if needed...
                processing(); // never ends
            default: // Parent
                quid_message(SYS, "New child process %d started", cpid);
        }     
    }
}


/**
 * main parses the command-line arguments and the config-file, 
 * calls qg_init to establish and verify connection to QualysGuard,
 * checks and sets all options and calls the (IDS)_main-function
 */
int main(int argc, char *argv[])
{
    DIR *cachedir_p = NULL;
    DIR *logdir_p = NULL;
    int optc;
    char goodfname[STR_BUF_LEN];
    char badfname[STR_BUF_LEN];
    char unfname[STR_BUF_LEN];

    char varbuf[STR_BUF_LEN];
    char logbuf[STR_BUF_LEN];
    char *dircheck = NULL;
    char progbuf[TFS_ARR_LEN];

    snprintf(progbuf, TFS_ARR_LEN, "QuIDScor %s", VERSION);
    program = strdup(progbuf);

    // Arg check
    if (argc < 2) {
        printf("Please use %s -h for help in using quidscor\n", argv[0]);
        quid_exit(QSCE_ARGS);
    }

    // Parse arguments
    while ((optc = getopt(argc, argv, "C:M:c:e:i:l:m:p:r:s:u:DVdh01")) != -1) {
        // order in switch() is the same as in getopt()
        switch (optc) {
                // Specify Cache Directory
            case 'C':
                cachedir = optarg;
                break;

                // Specify mode
            case 'M':
	        if (toupper(optarg[0]) == MODE_SNORT_C) {
                    mode = MODE_SNORT;
	        } else if (toupper(optarg[0]) == MODE_LOG_C) {
                    mode = MODE_LOG;
	        } else if (toupper(optarg[0]) == MODE_DRAGON_C) {
                    mode = MODE_DRAGON;
	        } else {
                    printf
                        ("Unknown mode: \"%s\" specified. Please select either \"snort\", or \"dragon\"\n",
                         optarg);
                    quid_exit(QSCE_MODE);
                }
                break;

                // Read from config file
            case 'c':
                parse_config(optarg);
                break;

                // Specify expiration time
            case 'e':
                expiretime = atoi(optarg);
                break;

                // Input (file/socket)
            case 'i':
                inputfname = optarg;
                break;

                // Specify Log Directory
            case 'l':
                logdir = optarg;
                break;

                // specify custom-mapping-file
            case 'm':
                mapfile = optarg;
                break;

                // Password
            case 'p':
                password = optarg;
                break;

                // Specify rules directory
            case 'r':
                rdir = optarg;
                break;

                // Specify qualysguard server
            case 's':
                qgserver = optarg;
                break;

                // Specify qualysguard username
            case 'u':
                username = optarg;
                break;

                // Daemon Mode
            case 'D':
                daemonmode = TRUE;
                break;

                // Show version information and exit
            case 'V':
                display_version();
                quid_exit(QSCE_NORMAL);
                break;

                // Debug mode
            case 'd':
                debug = TRUE;
                break;

                // DB create/refresh only
            case '0':
                dbRefreshOnly = TRUE;
                break;

                // Only use prexisting DB files
            case '1':
                dbReadOnly = TRUE;
                break;

	        // If illegal option go to help
	    default:
                // Help
            case 'h':
                display_version();
                printf("QuIDScor Help:\n");
                printf("  -c <file>  specifies a configuration file\n");
                printf("  -d         enable debugging mode.\n");
                printf("  -h         this help.\n");
                printf("  -D         run as a daemon.\n");
                printf("  -V         display version information and exit.\n");
                printf("  -0         update/create the VA and IDS db only (for use with -1)\n");
                printf("  -1         only run the correlation engine (no VA data fetching process)\n");
                printf("You can specify the following options (and more) in the configuration file:\n");
#ifdef QS_DRAGON
		printf("  -M <s|d|l> specifies which mode it is running in. (snort|dragon|log)\n");
#else
		printf("  -M <s|l>   specifies which mode it is running in. (snort socket|log)\n");
#endif
		printf("  -i <path>  input (source) for alerts (socket or log file)\n");
                printf("  -C <dir>   specifies cache file directory\n");
                printf("  -r <dir>   specifies a directory to read rules from (required)\n");
                printf("  -l <dir>   specifies the directory that will hold the logfiles.\n");
// Only needed in dev/qa; no point in documenting the option in the scarce
// one page online help real estate
//                printf("  -s <url>   specified the URL of the QualysGuard server.\n");
                printf("  -u <user>  specifies the QualysGuard username\n");
                printf("  -p <pwd>   specifies the QualysGuard password\n");
                printf("  -e <sec>   specifies the length of time (in seconds) that data is cached.\n");
                printf("  -m <file>  specifies a custom-mapping file like snortmap.xml\n");
                quid_exit(QSCE_ARGS);
                break;

        }
    }

    // Print Banner
    if (!daemonmode){
        display_version();
    }

    if (mapfile) {
        if (mapfile[0] != '/') {
            quid_message(ERR, "Custom mapping file must be an absolute path (\"%s\" given)", mapfile);
            quid_exit(QSCE_ARGS);
        }
    } else {
        printf("[*] Warning: no custom mapping specified. Features such as service and\n    application correlation will not be enabled.\n");
    }

    if (dbReadOnly && dbRefreshOnly) {
        quid_message(ERR, "-0 and -1 options are exclusive.");
        quid_exit(QSCE_ARGS);
    }

    if (dbRefreshOnly && daemonmode) {
        quid_message(ERR, "-0 (db refresh) and -D (daemon mode) are exclusive.");
        quid_exit(QSCE_ARGS);
    }

    if (expiretime < MIN_EXPIRE) {
        quid_message(ERR, "Expiration time set unreasonably low, please use >= %d seconds", MIN_EXPIRE);
        quid_exit(QSCE_ARGS);

    }

    if (cachedir == NULL) {
        snprintf(varbuf, STR_BUF_LEN, "%s/%s", VARPREFIX, CACHE_DIR);
        cachedir = strdup(varbuf);
    }

    if (logdir == NULL) {
        snprintf(logbuf, STR_BUF_LEN, "%s/%s", VARPREFIX, LOG_DIR);
        logdir = strdup(logbuf);
    }

    reform_path(cachedir);
    reform_path(logdir);
    reform_path(qgserver);

    // Make sure we've defined a mode
    if (mode == MODE_UNDEFINED) {
        printf("You must select which mode we're in via the -M flag.\n");
        quid_exit(QSCE_MODE_UNDEF);
    }

    if (inputfname == NULL) {
        if (mode == MODE_LOG) {
            inputfname = strdup(INPUT_LOGFNAME);
        } else {
            inputfname = strdup(SNORT_SOCKET);
        }
    }    


    // Did the user specify the rules dir?
    if (!rdir && !dbReadOnly) {
        printf
            ("ERROR: You must specify a directory where rules can be read in by using the -r option or RULESDIR in the config file.\n");
        quid_exit(QSCE_RULESDIR);
    }

    // Setup Signal Handlers
    signal(SIGTERM, sighandler);
    signal(SIGHUP, sighandler);
    signal(SIGPIPE, sighandler);
    signal(SIGINT, sighandler);
    signal(SIGUSR1, sighandler);
    // TODO: I recall this is needed, I think for either -1 or -0 mode
    // instead of directly registering sighandler - but why ?
    signal(SIGCHLD, SIG_IGN);

    // Open Syslog
    openlog("quidscor", 0, LOG_DAEMON);
    quid_message(SYS, "QuIDScor starting");
 
    quid_message(DBG, "[*] Cache DB:\t\t%s", cachedir);

    // Insure we can open the cache directory
    if ((cachedir_p = opendir(cachedir)) == NULL) {
        if ((dircheck = create_dir(cachedir)) != NULL) {
            printf
                ("Please insure that directory %s can be accessed or specify another directory for the cache using the -C option.\n",
                 dircheck);
            quid_exit(QSCE_CACHEDIR);
        }
    }
    if (cachedir_p != NULL) {
        closedir(cachedir_p);
    }

    if (dbRefreshOnly) {
        printf("[*] (Re)Building cache DB only mode\n");
    } else {
        quid_message(DBG, "[*] Logging to:\t\t%s", logdir);
        
        // Insure we can open the log directory
        if ((logdir_p = opendir(logdir)) == NULL) {
            if ((dircheck = create_dir(logdir)) != NULL) {
                printf
                    ("Please insure that directory %s can be accessed or specify another directory for the logs using the -l option.\n",
                     dircheck);
                quid_exit(QSCE_LOGDIR);
            }
        }
        if (logdir_p != NULL) {
            closedir(logdir_p);
        }
        // Build Logfile names
        snprintf(goodfname, STR_BUF_LEN, "%s/%s", logdir, AL_GOOD);
        snprintf(badfname, STR_BUF_LEN, "%s/%s", logdir, AL_BAD);
        snprintf(unfname, STR_BUF_LEN, "%s/%s", logdir, AL_UNVER);

        quid_message(DBG, "[*] Prioritizing to:\t%s\n\t\t\t%s\n\t\t\t%s", 
                     goodfname, unfname, badfname);

        // Open files
        if ((good = fopen(goodfname, "a+")) == NULL
            || (bad = fopen(badfname, "a+")) == NULL
            || (unverified = fopen(unfname, "a+")) == NULL) {
            perror("Couldn't open prioritized logfile(s) for alerts. Consider changing log directory using the -l option");
            quid_exit(QSCE_LOGFILE);
        }
    }

    // More Information
    if (!daemonmode) {
        if (!dbReadOnly) {
            printf("[*] Using QualysGuard server: %s\n", qgserver);
        }
        printf("[*] Cache DB expiry set to %lu seconds.\n", expiretime);
        quid_message(DBG, "[*] Debug Mode Enabled.");
    }

    if (dbReadOnly) {
        printf("[*] Using existing DB only mode\n");
    } else {
        // No username specified?
        if (!username) {
            printf
                ("ERROR: You must specify a username for the qualysguard server. Use -h for help.\n");
            quid_exit(QSCE_USERNAME);
        }

        // No password specified? Ask for it
        if (!password || (strlen(password) == 0)) {
            password = getpass("QualysGuard Password: ");
        }

        // Open QG Handle
        if ((handle = qg_init(username, password, qgserver, debug)) == NULL) {
            quid_message(ERR, strqg_error("qg_init failed"));
            quid_exit(QSCE_QGINIT);
        }
        reform_path(rdir);
    }

    // Load QG supported CVEs and custom mappings
    if (!load_mappings(handle, mapfile, rdir, mode, cachedir)) {
        quid_message(ERR, strqg_error("Error in mapping IDS-IDs with Qualys-IDs."));
        quid_exit(QSCE_MAP);
    }

    if (dbRefreshOnly) {
        printf("[*] DB refresh completed. Exiting.\n");
        exit(QSCE_NORMAL);
    }

    printf("[*] Initialization completed.\n");

    // Are we a daemon?
    switch (switch_to_daemon()) {
        case TRUE:
            //TRUE: daemon-option is deactivated or daemon is started
            break;
        case FALSE:
            //FALSE: daemon-option is deactivated, exit parent-process
            exit(QSCE_NORMAL);
        default:
            //default: error happened during start of child-process (daemon)
            quid_message(ERR, "Daemon could not be started.");
            quid_exit(QSCE_FORK);
    }            

    atexit(cleanup);

    if (daemonmode) {
        quid_message(SYS, "QuIDScor is now running in daemon-mode.");
    }

    // Split into 2 process only if the user didn't already request
    // half (-1/-0) and we are not in log processing mode (where 
    // async refresh of the DB does not make sense)
    // TODO: there's been report of hanging in log mode; try
    // again without that last part of the if
    // TODO: refactor with va_update_loop() above (and maybe switch_to_dameon()
    // too)
    if (!dbReadOnly && !dbRefreshOnly && (mode != MODE_LOG)) {
        // Default mode : split into 2 processes
        quid_message(SYS, "Splitting into 2 processes.");
        cpid = fork();
        switch (cpid) {
            case -1: // Error
                quid_message(ERR, "Split process could not be started.");
                quid_exit(QSCE_FORK);
            case 0 : // Child
                dbReadOnly = TRUE;
                break;
            default: // Parent
                signal(SIGCHLD, sighandler);
                quid_message(SYS, "Child process %d started", cpid);
                va_update_loop(); // never ends...
        }                
    }
    processing();
    return 0; // never reached
}


/**
 * parse_config loads the config-file, specified by 'file', and parses
 * it. parse_config quits QuIDScor if an error occurres during 
 * processing of the file.
 */
void parse_config(char *file)
{
    int i = 0;
    int argcount = 0;
    char buf[STR_BUF_LEN];
    char *args[MAX_TOKENS];
    char *tok = NULL;
    FILE *fhandle = NULL;
    exec_t *eptr = NULL;
    char firstc;
    static char *delim = CONFDELIM;
    
    if (!file) {
        printf("No config file specified via -c. Quitting.\n");
        quid_exit(QSCE_NOCONF);
    }
    // Open config file
    if ((fhandle = fopen(file, "r")) == NULL) {
        perror("Couldn't open config file");
        quid_exit(QSCE_OPENCONF);
    }
    // Parse file
    while (fgets(buf, STR_BUF_LEN, fhandle) != NULL) {
        // Clear out args
        for (i = 0; i < MAX_TOKENS; i++) {
            args[i] = NULL;
        }
        argcount = 0;

        // Parse: space delimited
        for (tok = strtok(buf, delim);
             (tok != NULL) && (argcount < MAX_TOKENS);
             tok = strtok(NULL, delim)) {
            // skip empty tokens (multiple spaces, end of line,...)
            if (strlen(tok)==0) {
                continue;
            }
            args[argcount++] = tok;
        }

        // skip empty lines and comments
        if ((argcount == 0) || (args[0][0] == SIGN_COMM)) {
            continue;
        }

        // Mode
        if (strcasecmp(args[0], CO_MOD) == 0) {
            if (argcount != 2) {
                printf("%s configuration takes only 1 argument, found %d\n",
                       CO_MOD, argcount - 1);
                quid_exit(QSCE_CONFILE);
            }
            firstc = toupper(args[1][0]);
            if (firstc == MODE_SNORT_C) {
                mode = MODE_SNORT;
            } else if (firstc == MODE_DRAGON_C) {
                mode = MODE_DRAGON;
            } else if (firstc == MODE_LOG_C) {
                mode = MODE_LOG;
            } else {
                printf("Unknown mode: %s\n", args[1]);
            }
        } else                  // Server
	if (strcasecmp(args[0], CO_SER) == 0) {
	    if (argcount > 4 || argcount < 3) {
 	        printf("%s needs at least 2 arguments (<1>server and <2>username, <3>password is optional), found %d\n", 
		       CO_SER, argcount - 1);
		quid_exit(QSCE_CONFILE);
	    }
            qgserver = strdup(args[1]);
            username = strdup(args[2]);
            if (argcount > 3) {
                password = strdup(args[3]);
            }
        } else                  // Logdir
	if (strcasecmp(args[0], CO_LOG) == 0) {
	    if (argcount != 2) {
	        printf("%s needs one argument, found %d\n",
	      	       CO_LOG, argcount - 1);
		quid_exit(QSCE_CONFILE);
	    }
            logdir = strdup(args[1]);
        } else                  // Cachedir
	if (strcasecmp(args[0], CO_CAC) == 0) {
	    if (argcount != 2) {
		printf("%s needs one argument, found %d\n",
		       CO_CAC, argcount - 1);
		quid_exit(QSCE_CONFILE);
	    }
            cachedir = strdup(args[1]);
        } else                  // Rules Dir
	if (strcasecmp(args[0], CO_RUL) == 0) { 
	    if (argcount != 2) {
		printf("%s needs one argument, found %d\n", 
		       CO_RUL, argcount - 1);
		quid_exit(QSCE_CONFILE);
	    }
            rdir = strdup(args[1]);
        } else                  // Expiretime
	if (strcasecmp(args[0], CO_EXP) == 0) {
	    if (argcount != 2) {
	        printf("%s needs one argument, found %d\n",
		       CO_EXP, argcount - 1);
		quid_exit(QSCE_CONFILE);
	    }
            expiretime = atoi(args[1]);
        } else                  // Custom mapping-file
        if (strcasecmp(args[0], CO_MAP) == 0) {
            if (argcount !=2) {
                printf("%s need one argument, found %d\n",
                       CO_MAP, argcount - 1);
                quid_exit(QSCE_CONFILE);
            }
            mapfile = strdup(args[1]);
        } else                  // Source = Snortsocket / input log file
        if (strcasecmp(args[0], CO_SRC) == 0) {
            if (argcount != 2) {
                printf("%s needs one argument, found %d\n",
                       CO_SRC, argcount - 1);
                quid_exit(QSCE_CONFILE);
            }
            inputfname = strdup(args[1]);
        } else                  // Exec
	if (strcasecmp(args[0], CO_EXE) == 0) {
	    if (argcount <= 4) {
	        printf("%s needs at least four arguments:\n\t<1> TRUE/FALSE for execution at \"good\" alerts\n\t<2> TRUE/FALSE for execution at \"bad\" alerts\n\t<3> TRUE/FALSE for execution at \"unverified\" events\n\t<4> command to be executed\n\tOther arguments are optional.\nActually found arguments for EXEC: %d\n", 
		       CO_EXE, argcount - 1);
		quid_exit(QSCE_CONFILE);
	    }
            if ((eptr = calloc(1, sizeof(exec_t))) == NULL) {
                perror("calloc error in config file parser");
                quid_exit(QSCE_MEMORY);
            }
            // Run for "good" events
            if (toupper(args[1][0]) == SIGN_TRUE) {
                eptr->good = 1;
	    } else {
                eptr->good = 0;
	    }

            // Run for "bad" events
            if (toupper(args[2][0]) == SIGN_TRUE) {
                eptr->bad = 1;
	    } else {
                eptr->bad = 0;
	    }

            // Run for "unverified" events
            if (toupper(args[3][0]) == SIGN_TRUE) {
                eptr->unverified = 1;
	    } else {
                eptr->unverified = 0;
	    }

            // Set number of args
            eptr->argc = (argcount - 4);

            // Copy all the args
            for (i = 0; i < eptr->argc; i++) {
                eptr->argv[i] = strdup(args[(i + 4)]);
	    }

            // Add to list
            eptr->next = edb;
            edb = eptr;
        } else                  // IpGroup
        if (strcasecmp(args[0], CO_IPG) == 0) {
            if (argcount != 2) {
                printf("%s needs one argument, found %d\n",
                       CO_IPG, argcount - 1);
                quid_exit(QSCE_CONFILE);
            }
            ipgroup = strdup(args[1]);
        } else {
            // unknown command/variable:
            printf("Unknown config entry \"%s\"\n",
                   args[0]);
            quid_exit(QSCE_CONFILE);
        }
    }

    fclose(fhandle);
    return;
}
