/*
 * $Id: quidscor_alert.c,v 1.54 2003/10/08 01:03:13 ldemailly Exp $
 *
 * Copyright (c) 2003, Qualys, Inc.  All rights reserved.
 *
 * See the file LICENSE for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */
#include "quidscor.h"

extern FILE *good;
extern FILE *bad;
extern FILE *unverified;
extern int debug;
extern int daemonmode;
extern idsdb_t idsdb;
extern exec_t *edb;

/**
 * process_alert processes the QuIDScor-alert-structure and calls
 * the correlation-function check-alert, after processing it 
 * frees the vhost_t-structure
 */
void process_alert(Alert *alert)
{
    // Print to stdout if in debug mode
    if (debug && !daemonmode) {
        print_alert(stdout, OUTPUT_ALERT, alert);
    }

    // Get rid of broadcast
    if (ntohl(alert->source_ip.s_addr) == 0
        || ntohl(alert->dest_ip.s_addr) == ADR_BROADCAST) {
        // TODO: use different sub category ?
        print_alert(unverified, VULN_NO_REPORT_FOUND, alert);
        return;
    }

#ifdef USE_XPATH
    // Find host entry
    if (!find_host(alert->dest_ip.s_addr, TRUE)) {
        print_alert(unverified, VULN_NO_REPORT_FOUND, alert);
        quid_message(DBG, "[E] Couldn't find report for host: %s",
                   inet_ntoa(alert->dest_ip));
        return;
    }

#else
    idsdb.reports = xmlHashLookup(idsdb.reporthash, inet_ntoa(alert->dest_ip));
    if (idsdb.reports == NULL) {
        print_alert(unverified, VULN_NO_REPORT_FOUND, alert);
        quid_message(DBG, "[E] Couldn't find report for host: %s",
                   inet_ntoa(alert->dest_ip));
        return;
    }
#endif
    check_alert(alert);

    free_alert(alert);

}

/**
 * check_alert correlates the alert from the IDS with the reports from
 * QualysGuard using the QuIDScor-alert-structure and the
 * vhost_t-structure which contain all relevant information
 */
void check_alert(Alert *alert) 
{
    char *class = NULL;
    char *alertcat = NULL;
    QG_pcheck hasPORT = -1;
    QG_detcheck hasSERVICE = -1;
    QG_detcheck hasAPPLIC = -1;

    alert->qids = findMapping(alert);

    hasPORT = check_Ports(alert);
    
    // Check the result of the checked ports
    switch (hasPORT) {
    case T_FOUND:
        if (!alert->qids) {
            alertcat = VULN_TCP_PORT_VERIFIED;
            hasSERVICE = check_Detail(alert, DET_SERV);
        }
        break;
    case T_ESTAB:
        print_alert(good, TCP_PORT_WARNING, alert);
        return;
    case T_NOPORT:
        print_alert(bad, VULN_NO_TCP_PORT_VERIFIED, alert);
        return;
    case U_FOUND:
        if (!alert->qids) {
            alertcat = VULN_UDP_PORT_VERIFIED;
            hasSERVICE = check_Detail(alert, DET_SERV);
        }
        break;
    case U_NOPORT:
        print_alert(bad, VULN_NO_UDP_PORT_VERIFIED, alert);
        return;
    default:
        if (!alert->qids) {
            print_alert(unverified, VULN_NO_VERIFICATION, alert);
            return;
        }
        break;
    }

    switch (hasSERVICE) {
    case DET_FOUND:
        hasAPPLIC = check_Detail(alert, DET_APP);
        if (hasPORT == T_FOUND) {
            alertcat = VULN_TCP_SERV_VERIFIED;
        } else {
            alertcat = VULN_UDP_SERV_VERIFIED;
        }
        break;
    case DET_NOTFOUND:
        if (hasPORT == T_FOUND) {
            alertcat = VULN_NO_TCP_SERV_VERIFIED;
        } else {
            alertcat = VULN_NO_UDP_SERV_VERIFIED;
        }
        print_alert(bad, alertcat, alert);
        return;
    case DET_NOMAP:
        print_alert(good, alertcat, alert);
        return;
    default:
        break;
    }
    
    switch (hasAPPLIC) {
    case DET_FOUND:
        if (hasPORT == T_FOUND) {
            alertcat = VULN_TCP_APP_VERIFIED;
        } else {
            alertcat = VULN_UDP_APP_VERIFIED;
        }
        print_alert(good, alertcat, alert);
        return;
    case DET_NOTFOUND:
        if (alert->prottype == IPPROTO_TCP) {
            alertcat = VULN_NO_TCP_APP_VERIFIED;
        } else {
            alertcat = VULN_NO_UDP_APP_VERIFIED;
        }
        print_alert(bad, alertcat, alert);
        return;
    case DET_NOMAP:
        print_alert(good, alertcat, alert);
        return;
    default:
        break;
    }

    // check if the report contains the matching CVE and port as Vuln or Threat
    if ((class = check_CVE(alert)) != NULL) {
        if (strcmp(class, "VULN") == 0) {
            print_alert(good, VULN_ONPORT_VERIFIED, alert);
        } else if (strcmp(class, "THREAT") == 0) {
            print_alert(good, THREAT_ONPORT_VERIFIED, alert);
        } else if (strcmp(class, "VULNnoport") == 0) {
            // free class because it was reallocated with qg_strappend
            print_alert(good, VULN_VERIFIED, alert);
        } else {
            // free class because it was reallocated with qg_strappend
            print_alert(good, THREAT_VERIFIED, alert);
        }
        xmlFree(class);
    } else {
        print_alert(bad, VULN_NO_VULN_VERIFIED, alert);
    }
    return;
}

/**
 * findMapping searches in the mapping structure for mappings and returns TRUE
 * if it has found mappings, otherwise FALSE
 */
int findMapping (Alert *alert) {
#ifdef USE_XPATH
    char sidbuf[TFS_ARR_LEN];
    xmlXPathObjectPtr idsptr = NULL;
    // search with XPATH
    snprintf(sidbuf, TFS_ARR_LEN, "/IDSTOQID/IN[@TYPE=\"ID\"][@VALUE=\"%s\"]/OUT", alert->rid);
    idsptr = xmlXPathEvalExpression(sidbuf, idsdb.idstoqid);
    
    if (xmlXPathNodeSetIsEmpty(idsptr->nodesetval)) {
        alert->maps = NULL;
    } else {
        alert->maps = idsptr->nodesetval->nodeTab[0];
    }

    xmlXPathFreeObject(idsptr);
#else
    // search in Hashtable
    alert->maps = xmlHashLookup(idsdb.idstoqidhash, alert->rid);
#endif

    // no mapping at all
    if (alert->maps == NULL) {
        return FALSE;
    }

    // no ID-mapping found
    return flagValueSet(alert->maps,MAP_STR_ID,NULL);
}

/**
 * check_Detail searches for mapped services or applications in the QuIDScor-
 * reports. The parameter "detail" specifies which detail should be checked 
 * (either service or application). The function returns a value of the type 
 * QG_detcheck which is either no mapping (DET_NOMAP), found (DET_FOUND) 
 * or not found (DET_NOTFOUND)
 */
QG_detcheck check_Detail(Alert *alert, QG_detcheck detail)
{
    char *attrib = NULL;
    char *repdetail = NULL;
    int mapping = 0;
    xmlNodePtr mapnode = NULL;

    switch (detail) {
    case DET_SERV:
        attrib = "SERVICE";
        repdetail = XPATH_PROT_SERV;
        break;
    case DET_APP:
        attrib = "APPLICATION";
        repdetail = XPATH_PROT_APP;
        break;
    default:
        return DET_NOMAP;
    }

    // check every detected entry
    for (mapnode = alert->maps; mapnode != NULL; mapnode = mapnode->next) {
        char *maptype = NULL;
        char *detailname = NULL;
        char *xpathbuf = NULL;
        char repbuf[TFS_ARR_LEN];
        int det_found = FALSE;
        xmlXPathObjectPtr detailrep = NULL;
        
        maptype = xmlGetProp(mapnode, MAP_STR_TYPE);
        if (maptype == NULL) {
            continue;
        }
        if (strcmp(maptype, attrib)) {
            free(maptype);
            continue;
        }

        if ((detailname = xmlGetProp(mapnode, MAP_STR_VAL)) == NULL) {
            free(maptype);
            continue;
        }
        
        mapping = TRUE;
        xpathbuf = strdup(alert->xpath);

        snprintf(repbuf, TFS_ARR_LEN, repdetail, detailname);
        xpathbuf = qg_strappend(xpathbuf, repbuf);

        detailrep = xmlXPathEvalExpression(xpathbuf, idsdb.reports);

        if (!xmlXPathNodeSetIsEmpty(detailrep->nodesetval)) {
            free(alert->xpath);
            alert->xpath = strdup(xpathbuf);
            det_found = TRUE;
        }
        xmlXPathFreeObject(detailrep);
        xmlFree(detailname);
        free(xpathbuf);
        free(maptype);
        if (det_found) {
            return DET_FOUND;
        }
    }

    if (mapping) {
        return DET_NOTFOUND;
    }

    return DET_NOMAP;
}


QG_pcheck check_Ports(Alert *alert)
{
    int hasPORT = (alert->prottype == IPPROTO_TCP) || 
        (alert->prottype == IPPROTO_UDP);
    char xpathport[TFS_ARR_LEN];
    xmlXPathObjectPtr evalbuf = NULL;

    if (!hasPORT) {
        return UT_NOPORT;
    }

    // TCP/UDP: check if the service and the port are in the report
    if (alert->prottype == IPPROTO_TCP) {
        snprintf(xpathport, TFS_ARR_LEN, XPATH_PROT_PORT, 
                 inet_ntoa(alert->dest_ip), "TCP", alert->dest_port);
        alert->xpath = strdup(xpathport);
        // searches for '/REPORT/RESULT[@IP="%s"]/INFOS/PORT[@PROTOCOL="TCP"][@VALUE="%d"]'
        evalbuf = xmlXPathEvalExpression(alert->xpath, idsdb.reports);
        // Does the report verify this service?

        if (!xmlXPathNodeSetIsEmpty(evalbuf->nodesetval)) {
            xmlXPathFreeObject(evalbuf);
            return T_FOUND;
        }
        xmlXPathFreeObject(evalbuf);

        // established?
        if (alert->maps) {
            if (flagValueSet(alert->maps,MAP_STR_FLAG,MAP_STR_ESTABLISHED)){
                return T_ESTAB;
            }
        }
        return T_NOPORT;
    } 

    //Here comes the UDP-part

    snprintf(xpathport, TFS_ARR_LEN, XPATH_PROT_PORT,  
             inet_ntoa(alert->dest_ip), "UDP", alert->dest_port);
    if (alert->xpath) {
        free(alert->xpath);
    }
    alert->xpath = strdup(xpathport);
    // searches for '/REPORT/RESULT[@IP="%s"]/INFOS/PORT[@PROTOCOL="UDP"][@VALUE="%d"]'
    evalbuf = xmlXPathEvalExpression(alert->xpath, idsdb.reports);

    // do the same for UDP as for TCP
    if (!xmlXPathNodeSetIsEmpty(evalbuf->nodesetval)) {
        xmlXPathFreeObject(evalbuf);
        return U_FOUND;
    }
    
    xmlXPathFreeObject(evalbuf);
    return U_NOPORT;
}


/**
 * check_CVE checks if the report of QualysGuard contains a matching Qualys-ID
 * at the matching port in the cvelist, if the protocol uses ports (TCP/UDP) 
 * and returns the classification of the alert -> Vulnerability / Threat / NULL
 */
char *check_CVE(Alert *alert)
{
    xmlNodePtr mapqid = alert->maps;
    char *class = NULL;
    char *mapstr = NULL;

    while (mapqid != NULL) {
        char *mapping = NULL;
        char xpathvuln[TFS_ARR_LEN];
        xmlXPathObjectPtr evalbuf = NULL;
        
        if ((mapping = xmlGetProp(mapqid, MAP_STR_TYPE)) == NULL) {
            mapqid = mapqid->next;
            continue;
        }

        if ((mapstr = xmlGetProp(mapqid, MAP_STR_VAL)) == NULL) {
            xmlFree(mapping);
            mapqid = mapqid->next;
            continue;
        }

        snprintf(xpathvuln, TFS_ARR_LEN, XPATH_VULN, inet_ntoa(alert->dest_ip), 
                 mapstr);
        // searches for '/REPORT/RESULT[@IP="%s"]/VULNS/FOUND[@ID=\"%s\"]'
        evalbuf = xmlXPathEvalExpression(xpathvuln, idsdb.reports);

        if (xmlXPathNodeSetIsEmpty(evalbuf->nodesetval)) {
            // got to next mapped-id
            xmlFree(mapping);
            xmlFree(mapstr);
            mapstr = NULL;
            xmlXPathFreeObject(evalbuf);
            mapqid = mapqid->next;
            continue;
        }

        // if network protocol uses ports, check for port
        if (alert->prottype == IPPROTO_TCP || alert->prottype == IPPROTO_UDP) {
            char *portstr = NULL;
            if ((portstr = xmlGetProp(evalbuf->nodesetval->nodeTab[0], "PORT")) 
                != NULL) {
                if (atoi(portstr) != alert->dest_port) {
                    xmlFree(mapping);
                    xmlFree(mapstr);
                    mapstr = NULL;
                    xmlFree(portstr);
                    xmlXPathFreeObject(evalbuf);
                    mapqid = mapqid->next;
                    continue;
                }
                xmlFree(portstr);
                class = xmlGetProp(evalbuf->nodesetval->nodeTab[0], "TYPE");
            }
        }

        // creates the string for assigning alert to classification if no port
        // was used or found
        if (class == NULL) {
            class = xmlGetProp(evalbuf->nodesetval->nodeTab[0], "TYPE");
            class = qg_strappend(class, "noport");
        }

        // saves the Qualys-ID and CVE-ID in the output
        alert->quid = strdup(mapstr);
        alert->cve = xmlGetProp(mapqid->children, "ID");

        xmlXPathFreeObject(evalbuf);
        xmlFree(mapstr);
        xmlFree(mapping);
        return class;
    }
    
    return class;
}

/**
 * flagValueSet checks in XML-tags for TYPE and VALUE and their corresponding values
 * The parameter 'value' is optional and if set to NULL flagValueSet only checks
 * for a matching TYPE
 */
int flagValueSet (xmlNodePtr node, char* type, char* value) {
    xmlNodePtr idsptr = node;

    if (type == NULL) {
        return FALSE;
    }

    for (; idsptr != NULL; idsptr = idsptr->next) {
        char *typeStr = NULL;
        char *valueStr = NULL;

        typeStr = xmlGetProp(idsptr, MAP_STR_TYPE);
                
        if (typeStr == NULL) {
            continue;
        }

        if (strcmp(type, typeStr)) {
            xmlFree(typeStr);
            continue;
        }

        if (value == NULL) {
            xmlFree(typeStr);
            return TRUE;
        }

        valueStr = xmlGetProp(idsptr, MAP_STR_VAL);

        if (valueStr == NULL) {
            xmlFree(typeStr);
            continue;
        }

        if (strcmp(valueStr,value)){
            xmlFree(valueStr);
            xmlFree(typeStr);
            continue;
        }

        xmlFree(valueStr);
        xmlFree(typeStr);
        return TRUE;
    }

    return FALSE;
}

/**
 * print_alert prints the alert to the output specified by 'fhandle'
 * 'verreas' stands for verification-reason und should contain a
 * char describing if it was verified or not
 */
void print_alert(FILE * fhandle, char * verreas, Alert * alert)
{
    char *quidstr = NULL;
    char *cvestr = NULL;
    struct protoent *pprot;

    pprot = getprotobynumber(alert->prottype);
    if (pprot == NULL) {
        quid_message(ERR, "Unknown protocol #%ld", alert->prottype);
    } else {
        alert->proto = pprot->p_name;
    }

    // Print to file
    fprintf(fhandle, "%s; ", alert->timestr);
    fprintf(fhandle, "%s; %s; %s; %s; %s; %s; ", 
            verreas, alert->rid, alert->alertmsg, alert->alertclass, 
            alert->additional, alert->proto);
    fprintf(fhandle, "%s:%d; ", 
            inet_ntoa(alert->source_ip), alert->source_port);
    fprintf(fhandle, "%s:%d;", inet_ntoa(alert->dest_ip), alert->dest_port);

    if ((quidstr = alert->quid) != NULL) {
        fprintf(fhandle, " Qualys-ID: %s;", quidstr);
    }

    if ((cvestr = alert->cve) != NULL) {
        fprintf(fhandle, " CVE/CAN: %s\n", cvestr);
    } else {
        fprintf(fhandle, "\n");
    }

    fflush(fhandle);

    //check for exec-list
    if (edb == NULL) {
        return;
    }

    // Do exec
    if (fhandle == good) {
        alert_exec(ALERT_GOOD, alert);
    }
    else if (fhandle == bad) {
        alert_exec(ALERT_BAD, alert);
    }
    else if (fhandle == unverified) {
        alert_exec(ALERT_UNVERIFIED, alert);
    }
}


// Experimental (Alpha) exec feature - TODO: audit/improve

/**
 * alert_exec executes the in the config-file specified commands 
 */
void alert_exec(u_int type, Alert *alert)
{
    int i = 0;
    char addit[TFS_ARR_LEN];
    char class[TFS_ARR_LEN];
    char alertmsg[STR_BUF_LEN];
    char sourceip[TFS_ARR_LEN];
    char destip[TFS_ARR_LEN];
    char sourceport[SHO_ARR_LEN];
    char destport[SHO_ARR_LEN];
    char cvstr[TT_ARR_LEN];
    char fullstr[STR_BUF_LEN];
    exec_t *eptr = NULL;
    char *eargs[SHO_ARR_LEN];
    char *exstr = NULL;

    if (!alert || edb == NULL) {
        quid_message(DBG, "No commands to execute specified.");
        return;
    }

    snprintf(sourceip, TFS_ARR_LEN, "%s", inet_ntoa(alert->source_ip));
    snprintf(destip, TFS_ARR_LEN, "%s", inet_ntoa(alert->dest_ip));
    snprintf(sourceport, SHO_ARR_LEN, "%d", alert->source_port);
    snprintf(destport, SHO_ARR_LEN, "%d", alert->dest_port);
    snprintf(class, TFS_ARR_LEN, "%s", alert->alertclass);
    snprintf(alertmsg, STR_BUF_LEN, "%s", alert->alertmsg);
    snprintf(addit, TFS_ARR_LEN, "%s", alert->additional);

    if (alert->cve != NULL) {
        snprintf(cvstr, SHO_ARR_LEN, alert->cve);
    } else {
        snprintf(cvstr, SHO_ARR_LEN, "None");
    }

    // Build fullstr
    snprintf(fullstr, STR_BUF_LEN,
             "\"\"%s\"; Classification: %s; Type: %s; %s:%s -> %s:%s; %s; CVE: %s;\"",
             alertmsg, class, alert->proto, sourceip, sourceport, destip, destport, addit, cvstr);


    // Do exec stuff
    for (eptr = edb; eptr != NULL; eptr = eptr->next) {
        // Insure this is an exec that matches this alert type
        if ((type == ALERT_GOOD && !eptr->good)
            || (type == ALERT_BAD && !eptr->bad)
            || (type == ALERT_UNVERIFIED && !eptr->unverified)) {
            continue;
        }

        // Process Args
        exstr = strdup(" ");
        for (i = 0; i < eptr->argc; i++) {
            // Do token replacement
            if (strcmp(eptr->argv[i], TO_ALER) == 0) {
                exstr = qg_strappend(exstr, fullstr);
            }                // CVE String
            else if (strcmp(eptr->argv[i], TO_CVE) == 0) {
                exstr = qg_strappend(exstr, cvstr);
            }                // Class
            else if (strcmp(eptr->argv[i], TO_ADDI) == 0) {
                exstr = qg_strappend(exstr, addit);
            }                // Priority
            else if (strcmp(eptr->argv[i], TO_CLAS) == 0) {
                exstr = qg_strappend(exstr, class);
            }                // Snort msg
            else if (strcmp(eptr->argv[i], TO_ALAR) == 0) {
                exstr = qg_strappend(exstr, alertmsg);
            }                // Full Alert
            else if (strcmp(eptr->argv[i], TO_PROT) == 0) {
                exstr = qg_strappend(exstr, alert->proto);
            }
            else if (strcmp(eptr->argv[i], TO_CLIE) == 0) {
                exstr = qg_strappend(exstr, sourceip);
            }                // Server IP
            else if (strcmp(eptr->argv[i], TO_SERV) == 0) {
                exstr = qg_strappend(exstr, destip);
            }                // Client port
            else if (strcmp(eptr->argv[i], TO_CLPO) == 0) {
                exstr = qg_strappend(exstr, sourceport);
            }                // Server port
            else if (strcmp(eptr->argv[i], TO_SEPO) == 0) {
                exstr = qg_strappend(exstr, destport);
            }
            else {
                exstr = qg_strappend(exstr, eptr->argv[i]);
            }
            exstr = qg_strappend(exstr, TO_EMPT);
        }

        eargs[0] = SHELL_EXEC;
        eargs[1] = SHELL_EXEC_OPT;
        eargs[2] = exstr;
        eargs[3] = NULL;

        // Run it
        if (fork() == 0) {
            // an execv version of the system() call. We do this
            // In order to support stuff like > & >>
            execv(eargs[0], eargs);
            exit(QSCE_NORMAL);
        } else {
            free(exstr);
        }
    }
}

void free_alert(Alert *alert) 
{
    if (alert->cve) {
        xmlFree(alert->cve);
    }

    if (alert->quid) {
        free(alert->quid);
    }

    if (alert->xpath) {
        free(alert->xpath);
    }
}
