/*
 * $Id: quidscor_cache.c,v 1.44 2003/10/06 23:33:55 ldemailly Exp $
 *
 * Copyright (c) 2003, Qualys, Inc.  All rights reserved.
 *
 * See the file LICENSE for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */
#include "quidscor.h"

extern qghandle_t *handle;
extern char *cachedir;
extern idsdb_t idsdb;
extern u_long expiretime;
extern int dbReadOnly;


/**
 * returns NULL if cache file is outdated or does not exist,
 * with -1 it never expires.
 * returns the file stat structure otherwise. (NOT THREAD SAFE).
 */
struct stat *need_update(char *cfname, u_long pexpiretime) {
    static struct stat statbuf;
    int ret;
    u_long age;
    ret = stat(cfname, &statbuf);
    if (ret != 0) {
        quid_message(DBG, "[D] Stat of \"%s\" failed with %d", cfname, ret);
        return NULL;
    }
    age = (time(NULL) - statbuf.st_mtime);
    quid_message(DBG, "[D] Successful stat of \"%s\" age=%ld, mtime=%ld", cfname, 
                 age, statbuf.st_mtime);
    if ((pexpiretime != -1) && (age > pexpiretime)) {
        quid_message(DBG, "[D] Outdated cache for \"%s\"", cfname);
        return NULL;
    }
    quid_message(DBG, "[D] Fresh cache for \"%s\"", cfname);
    return &statbuf;
}

/**
 * find_host searches for the report in the cache-directory which 
 * matches to the IP-address specified by 'ip'. If a report 
 * exists it checks if it is older than it should be. (The
 * expiration time can be specified in the config-file) If it is
 * too old, find_host fetches with the help of libqg the latest
 * report from QualysGuard.
 * returns NULL on error.
 */
int find_host(u_long ip, int refreshDB) {
    struct stat *statbuf;
    char cfname[STR_BUF_LEN], addrname[SHO_ARR_LEN];
    struct in_addr addr;
    u_long pexpiretime = (dbReadOnly ? -1 : expiretime);

    // Make addr string
    addr.s_addr = ip;
    snprintf(addrname, SHO_ARR_LEN, "%s", inet_ntoa(addr));

    // Make filename
    snprintf(cfname, STR_BUF_LEN, "%s/%s", cachedir, addrname);

    // Does the file exist? If it does is it too old? : trigger
    // point to decide to refresh the whole memory DB
    // (todo: incrementally replace the part of the tree that needs
    //        update instead)
    if ((statbuf = need_update(cfname, pexpiretime)) != NULL) {
        if (statbuf->st_size == 0) {
            quid_message(DBG, "[D] Empty report file \"%s\"", cfname);   
            return FALSE;
        }
        return TRUE;
    } else {
        int ret;
        if (dbReadOnly) {
            return FALSE;
        }
        ret = refresh_host(addrname, cfname);
        if (refreshDB) {
            qualys_reports(); // refresh the report DB
        }
        return ret;
    }
}

/**
 *
 * Returns a TRUE if found without error.
 */
int refresh_host(char *addrname, char *cfname) {
    xmlDocPtr custom = NULL;
    char *ref = NULL;
    FILE *fhandle = NULL;
    qgresponse_t *res = NULL;
    u_long size = 0;

    quid_message(DBG, "[D] Fetching Report for Host: %s", addrname);

    // Get report list
    if ((res = qg_scan_report_list(handle, 1, addrname)) == NULL
        || !res->xml) {
        qg_free_response(res);
        quid_message(ERR, strqg_error("No results"));
        return FALSE;
    }
    // We were now able to get data (possibly empty) about this host,
    // create the cached file:

    // Open file handle
    if ((fhandle = fopen(cfname, FMODE_WRITE)) == NULL) {
        quid_message(ERR, "Error, couldn't save cached data to file: %s:%s",
                     cfname, strerror(errno));
        qg_free_response(res);
        return FALSE;
    }
    quid_message(DBG, "[D] Creating cache entry for Host: %s", addrname);
        
    // Fetch latest report, and return NULL on error or NOT FOUND
    if ((ref = qg_xml_get_report_ref(res->xml)) == NULL || 
        strncmp(ref, LIBQG_NOT_FOUND, LIBQG_NFC_CONST) == 0) {
        if (ref == NULL) {
            // Error: print error message
            quid_message(ERR, strqg_error("Problem fetching report data."));
        }
        qg_free_response(res);
        fclose(fhandle);
        return FALSE;
    }

    // Free original response
    qg_free_response(res);

    // Get scan report
    if ((res = qg_scan_report(handle, ref, addrname)) == NULL
        || !res->xml) {
        if (res == NULL) {
            quid_message(ERR, strqg_error("scan_report"));
        }
        xmlFree(ref);
        qg_free_response(res);
        fclose(fhandle);
        return FALSE;
    }

    custom = qg_custom_report(res->xml, addrname, cachedir);
    // Write data to disk
    xmlDocDump(fhandle, custom);
    fclose(fhandle);

    // We have data now
    size = res->size;
    xmlFree(ref);
    qg_free_response(res);
    xmlFreeDoc(custom);
    // Null entry? Return!
    if (size == 0) {
        return FALSE;
    }

    return TRUE;
}

void qualys_reports()
{
    DIR *dp = NULL;
    struct dirent *direntry = NULL;
    struct stat statbuf;

#ifdef USE_XPATH
    xmlDocPtr allrep = NULL;
    xmlNodePtr allroot = NULL;
#endif    

    quid_message(DBG, "Reloading reports DB");

#ifdef USE_XPATH
    if (idsdb.reports != NULL) {
        xmlFreeDoc(idsdb.reports->doc);
        xmlXPathFreeContext(idsdb.reports);
    }
#else
    if (!idsdb.reporthash) {
        idsdb.reporthash = xmlHashCreate(MAX_IPS);
    }
#endif

    dp = opendir(cachedir);
    if (dp == NULL) {
        quid_message(ERR, "could not access directory \"%s\"", cachedir);
        exit(QSCE_CACHEDIR);
    }

#ifdef USE_XPATH
    allrep = xmlNewDoc("1.0");
    allroot = xmlNewDocNode(allrep, NULL, "REPORT", NULL);
    xmlDocSetRootElement(allrep, allroot);
#endif

    // walk through directory
    while ((direntry = readdir(dp))) {
        char *filename = strdup(cachedir);
        char *ipstr = NULL;
        xmlDocPtr report = NULL;
        xmlNodePtr reproot = NULL;
#ifdef USE_XPATH
        xmlNodePtr repcpy = NULL;
#else
        xmlXPathContextPtr reportcontextptr = NULL;
#endif

        // create filenames
        filename = qg_strappend(filename, "/");
        filename = qg_strappend(filename, direntry->d_name);
        
        // if the file statistics can not be created or the file is empty,
        // the file is empty or the directory-entry is no file next entry
        if (stat(filename, &statbuf) != 0 || 
            !S_ISREG(statbuf.st_mode) ||
            statbuf.st_size == 0) {
            free(filename);
            continue;
        }

        if ((report = xmlParseFile(filename)) == NULL) {
            quid_message(ERR, "No XML-file: %s", filename);
            free(filename);
            continue;
        }
        
        // if the parsed file is no report go to next directory-entry 
        if ((reproot = qg_xmlroot(report, "RESULT")) == NULL || 
            ((ipstr = xmlGetProp(reproot, "IP")) == NULL || 
             (strcmp(ipstr, direntry->d_name) != 0))) {
            free(filename);

            if (ipstr) {
                xmlFree(ipstr);
            }

            xmlFreeDoc(report);
            continue;
        }
        
#ifdef USE_XPATH
        // copy the report into the xml-structure, which should contain 
        //all reports
        repcpy = xmlCopyNode(reproot, 1);

        xmlAddChild(allroot, repcpy);

        xmlFreeDoc(report);
#else
        reportcontextptr = xmlHashLookup(idsdb.reporthash, ipstr);
        if (reportcontextptr) {
            xmlFreeDoc(reportcontextptr->doc);
            reportcontextptr->doc = report;
        } else {
            reportcontextptr = xmlXPathNewContext(report);
            xmlHashAddEntry(idsdb.reporthash, ipstr, reportcontextptr);
        }
#endif
        xmlFree(ipstr);
        free(filename);
    }
    
    closedir(dp);

#ifdef USE_XPATH
    if (idsdb.reports) {
        xmlXPathFreeContext(idsdb.reports);
    }
    idsdb.reports = xmlXPathNewContext(allrep);
#endif
}
