/*
 * $Id: quidscor_log_input.c,v 1.17 2003/10/08 01:03:13 ldemailly Exp $
 *
 * Copyright (c) 2003, Qualys, Inc.  All rights reserved.
 *
 * See the file LICENSE for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */
#include "quidscor.h"

extern int debug;
extern int daemonmode;

#define REGEX "^([-0-9:./ ]+)( +\\[\\*\\*\\]|;) +([A-Z-]+|\\[[^:]+)[:; ]+([^:;]+)(:[0-9]+\\]|;) +([^;]*)(;| +\\[\\*\\*\\]( +\\[Classification:)?)( +([^];]+)(;|\\]) +)?((\\[)?Priority: +([^];]+)(;|\\]))?( +\\{?([^;}]+)(\\}|;) +([0-9]+)\\.([0-9]+)\\.([0-9]+)\\.([0-9]+)(:([0-9]+))?(;| ->) ([0-9]+)\\.([0-9]+)\\.([0-9]+)\\.([0-9]+)(:([0-9]+))?;?( Qualys-ID: ([0-9]+);)?( CVE/CAN: (.*))?)?$"
// One more part as () groups as the whole exp is the first match.
#define REG_NUM_PARTS 38
// Time is first component
#define R_TIME 1
// IDS Id
#define R_ID 4
// Description
#define R_DESC 6
// Classification (optional)
#define R_CLAS 10
// Priority (optional)
#define R_PRIO 14
// Protocol (proto+ip+... all optional because of port scan alerts)
#define R_PROT 17
// Source ip start (+0-3)
#define R_SRC_IP 19
// Source port
#define R_SRC_PORT 24
// Dest ip start (+0-3)
#define R_DST_IP 26
// Dest port
#define R_DST_PORT 31

// When not specified, use:
#define NA "n/a"

/**
 * log_input_main is the main function for offline correlation. 
 * it opens log file and parses it for alerts
 */
void log_input_main(char *logfname)
{
    char  line_buf[STR_BUF_LEN];
    FILE* inputlog = NULL;
    int   ret;
    regex_t reg;

    if (strcmp(logfname,"-") == 0) {
        quid_message(DBG, "Using stdin for log input processing");
        inputlog = stdin;
    } else {
        inputlog = fopen(logfname, "r");
    }

    if (inputlog == NULL) {
        quid_message(ERR, "Error opening \"%s\": %s", logfname, 
                     strerror(errno));
        quid_exit(QSCE_SOURCE);
    }
    quid_message(DBG, "Compiling \"%s\", will extract %d components", REGEX, 
                 REG_NUM_PARTS);
    ret = regcomp(&reg, REGEX, REG_EXTENDED);
    if (ret) {
        regerror(ret, &reg, line_buf, STR_BUF_LEN);
        quid_message(ERR, "Regular expression error %d: %s", ret, line_buf);
        quid_exit(QSCE_REGEX);
    }
    while (fgets(line_buf, STR_BUF_LEN, inputlog)) {
        log_input_process_alert(&reg, line_buf);
    }
    printf("[*] Done processing all offline input.\n");
    regfree(&reg);
    quid_exit(QSCE_NORMAL);
}

char *extract_string(regmatch_t *pmatch, char *buf, char *defval) {
    // no match, return defaults
    if (pmatch->rm_so == -1) {
        return defval;
    }
    return buf + pmatch->rm_so;
}

int extract_int(regmatch_t *pmatch, char *buf, int defval) {
    // no match, return defaults
    if (pmatch->rm_so == -1) {
        return defval;
    }
    return atoi(buf + pmatch->rm_so);
}

unsigned long int extract_ip(regmatch_t *pmatch, char *buf) {
    int i;
    unsigned long int res = 0;
    for (i=0; i<4; i++) {
        res = (res << 8) + (extract_int(pmatch + i, buf, 0)) ;
    }
    return htonl(res);
}

/**
 * log_input_process_alert converts the log file structure to 
 * the common QuIDScor-alert structure (which allows multiple input
 * plugins / multiple IDS support)
 */
void log_input_process_alert(regex_t *preg, char *buf)
{
    Alert *alertp = NULL; 
    int len = strlen(buf);
    regmatch_t matches[REG_NUM_PARTS];
    int ret;
    char msg_buf[STR_BUF_LEN];
    int i;
    char *prio_str;
    struct protoent *pprot;

    while ((len > 0) && (
                  (buf[len-1]=='\n') 
               || (buf[len-1]=='\r')
               || (buf[len-1]=='\t')
               || (buf[len-1]==' ') )) {
        len--;
        buf[len]= '\0';
    }

    if (len < 1) {
        quid_message(DBG, "Empty line.");
        return; 
    }
    quid_message(DBG, "Read:'%s'", buf);
    ret = regexec(preg, buf, REG_NUM_PARTS, matches, 0 /* default flags */);
    if (ret != 0) {
        regerror(ret, preg, msg_buf, STR_BUF_LEN);
        quid_message(ERR, "Regex exec error %d: %s for alert \"%s\"", ret, msg_buf, buf);
        return;
    }
    for (i=1; i < REG_NUM_PARTS; i++) {
        int eo = matches[i].rm_eo;
        if (eo != -1) {
            buf[eo] = 0; // null terminate this section
        }
        // quid_message(DBG, "Regex match part %d : %d %d", i, matches[i].rm_so, eo);
    }
    alertp = calloc(1, sizeof(Alert));
    // Time
    alertp->timestr  = buf + matches[R_TIME].rm_so;
    // IDS (rule) id
    alertp->rid = buf + matches[R_ID].rm_so;
    // Title/Description/Msg
    alertp->alertmsg = buf + matches[R_DESC].rm_so;
    // Classification
    alertp->alertclass = extract_string(matches + R_CLAS, buf, NA);
    // Priority (optional)
    prio_str = extract_string(matches + R_PRIO, buf, NA);
    // Protocol (now optional group with the remaining data)
    if (matches[R_PROT].rm_so != -1) {
        alertp->proto = buf + matches[R_PROT].rm_so;
        pprot = getprotobyname(alertp->proto);
        if (pprot == NULL) {
            quid_message(ERR, "Unknown protocol \"%s\"", alertp->proto);
        } else {
            alertp->prottype = pprot->p_proto;
        }
    } else {
        alertp->prottype = 0; // 'IP' to avoid errors later
    }
    // Start IP (optional (will be 0 if not specified))
    alertp->source_ip.s_addr = extract_ip(matches + R_SRC_IP, buf);
    // Start port (optional)
    alertp->source_port = (u_short) extract_int(matches + R_SRC_PORT, buf, 0);
    // Dest IP (optional (will be 0 if not specified))
    alertp->dest_ip.s_addr = extract_ip(matches + R_DST_IP, buf);
    // Dest port (optional)
    alertp->dest_port = (u_short) extract_int(matches + R_DST_PORT, buf, 0);


    snprintf(msg_buf, STR_BUF_LEN, "Priority: %s", prio_str);
    alertp->additional = msg_buf;

    process_alert(alertp);

    // Free memory
    free(alertp);
}


