/*
 * $Id: quidscor_snort.c,v 1.54 2003/10/07 17:41:28 ldemailly Exp $
 *
 * Copyright (c) 2003, Qualys, Inc.  All rights reserved.
 *
 * See the file LICENSE for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */
#include "quidscor.h"

extern FILE *good;
extern FILE *bad;
extern FILE *unverified;
extern int debug;
extern int daemonmode;
extern idsdb_t idsdb;
extern int lastsig;


/**
 * snort_main is the main function for correlation with snort
 * it opens an UNIX-socket and listens for alerts from snort
 * if an alert is received snort_process_alert is called
 */
void snort_main(char *fname)
{
    int usock;
    int rlen;
    int usize = sizeof(struct sockaddr_un);
    int bufsize = sizeof(Snortpkt);
    struct sockaddr_un uaddr;
    char buf[bufsize];

    // Open UNIX Domain Socket
    if ((usock = socket(AF_UNIX, SOCK_DGRAM, 0)) == -1) {
        quid_message(ERR, "socket error");
        quid_exit(QSCE_SOCKET);
    }
    // Setup our AF_UNIX sockaddr
    memset(&uaddr, 0, sizeof(uaddr));

    fcntl(usock, F_SETFL, O_APPEND);
#ifndef NO_FCNTL_O_ASYNC
    fcntl(usock, F_SETFL, O_ASYNC);
#endif

    uaddr.sun_family = AF_UNIX;
    bcopy(fname, uaddr.sun_path, strlen(fname));
    // Need to reset/unlink to avoid "Address already in use" error
    unlink(fname);              // ignore ret/code error (only bind matters)
    // Create new snort AF_UNIX socket file
    if (bind(usock, (struct sockaddr *) &uaddr, usize) == -1) {
        quid_message(ERR, "bind error on \"%s\"", fname);
        quid_exit(QSCE_BIND);
    }
    
    quid_message(DBG, "[*] QuIDScor starts listening");

    // Main loop
    while (1) {
        // Did we recieve a message? Process it
        if ((rlen =
             recvfrom(usock, &buf, bufsize, (int) NULL,
                      (struct sockaddr *) &uaddr, &usize)) > 0) {
            // Call alert processer
            snort_process_alert(buf, rlen);
        }
        if (lastsig != 0) {
            quid_message(SYS, "Signal %d -> stopping work.", lastsig);
            break;  // We've been signaled to exit gracefully
        }
        if (rlen <= 0) {
            quid_message(ERR, "unexpected error in recvfrom (rlen=%d)", rlen);
        }
    }

    quid_exit(QSCE_NORMAL);
}

/**
 * snort_process_alert converts the snort-alert-structure to 
 * QuIDScor-alert structure which would allow easy implementation
 * of correlation with other IDS
 */
void snort_process_alert(char *buf, size_t len)
{
    struct ip iph_data;
    struct ip *iph_p = NULL;
    struct tcphdr tcp_data;
    struct tcphdr *tcp_p = NULL;
    struct udphdr udp_data;
    struct udphdr *udp_p = NULL;
    char alertbuf[STR_BUF_LEN];
    char addbuf[STR_BUF_LEN];
    char rid[SHO_ARR_LEN];    // Rule ID
    char timestr[SHO_ARR_LEN];
    struct tm btime;
    struct tm *btime_p = NULL;
    u_short sport = 0, dport = 0;
    Snortpkt *snort ;
    Alert alert;

    // Sanity Check
    if (!buf || len < 1) {
        quid_message(ERR, "No valid alert from snort.");
        return;
    }

    // 0/Null initialize the alert structure
    memset(&alert, 0, sizeof(Alert));

    // Setup ptr to alert data
    snort = (Snortpkt *) buf;
    // Determine suspected Vuln IP/Port

    iph_p = &iph_data;
    //memcpy to prevent error on system with pointer-alignment
    //ie: Solaris: pointers have to be dividable by 4
    memcpy(iph_p, snort->pkt + snort->nethdr, sizeof(iph_data));

    snprintf(alertbuf, STR_BUF_LEN, "%d", snort->event.classification);
    alert.alertclass = strdup(alertbuf);
    alert.alertmsg = snort->alertmsg;

    // Protocoltype
    alert.prottype = iph_p->ip_p;
    
    // IP-addresses
    alert.source_ip = iph_p->ip_src;
    alert.dest_ip = iph_p->ip_dst;

    // convert time in seconds to time-string
    btime_p = &btime;
    btime_p = localtime((time_t *)&snort->ts.tv_sec);
    strftime(timestr, SHO_ARR_LEN, DATE_FORMAT, btime_p);
    alert.timestr = strdup(timestr);

    // ports
    switch (iph_p->ip_p) {
        case IPPROTO_TCP:
            tcp_p = &tcp_data;
            memcpy(tcp_p, snort->pkt + snort->transhdr, sizeof(tcp_data));
            sport = tcp_p->th_sport;
            dport = tcp_p->th_dport;
            alert.source_port = ntohs(sport);
            alert.dest_port = ntohs(dport);
            break;
        case IPPROTO_UDP:
            udp_p = &udp_data;
            memcpy(udp_p, snort->pkt + snort->transhdr, sizeof(udp_data));
            sport = udp_p->uh_sport;
            dport = udp_p->uh_dport;
            alert.source_port = ntohs(sport);
            alert.dest_port = ntohs(dport);
            break;
        default:
            alert.source_port = 0;
            alert.dest_port = 0;
            break;
    }

    // Lookup any Snort-IDs
    snprintf(rid, SHO_ARR_LEN, "%d", snort->event.sig_id); 
    alert.rid = rid;

    alert.cve = NULL;
    alert.quid = NULL;

    // additional options IDS-specific
    snprintf(addbuf, STR_BUF_LEN, "Priority: %d", snort->event.priority);
    alert.additional = strdup(addbuf);
    
    process_alert(&alert);

    // free pointers
    free(alert.alertclass);
    free(alert.timestr);
    free(alert.additional);
}

/**
 * snort_load_alertfile loads pairs of Snort-IDs and CVEs 
 * which is necessary for correlating 
 */
void snort_load_alertfile(char *file, xmlDocPtr siddb)
{
    char buf[STR_BUF_LEN];
    FILE *fhandle = NULL;
    xmlNodePtr sidroot = NULL;
    xmlNodePtr estroot = NULL;

    // Open file
    if ((fhandle = fopen(file, FMODE_READ)) == NULL) {
        snprintf(buf, STR_BUF_LEN, "Can't load sigs from \"%s\"", file);
        perror(buf);
        quid_exit(QSCE_SIG_FILE);
    }

    quid_message(DBG, "Loading CVE's from: %s", file);

    sidroot = xmlDocGetRootElement(siddb);
    estroot = xmlDocGetRootElement(idsdb.idstoqid->doc);

    while (fgets(buf, STR_BUF_LEN, fhandle) != NULL) {
        xmlNodePtr idsInPtr = NULL;
        char sigbuf[SHO_ARR_LEN];
        char *cvesearch = NULL;
        char *sidsearch = NULL;
        char *cvefound = NULL;
        char *established = NULL;
        int cvestart = idsdb.cvedbsize;

        // TODO: tokenize by ; and , instead of relying on hardcoded
        // string length and full string search
        if ((sidsearch = strstr(buf, SIG_SID)) == NULL) {
            continue;
        }

        // Skip "sid:"
        sidsearch += SIDID_OFFSET;

        // Fill in entry
        snprintf(sigbuf, SHO_ARR_LEN, "%d", atoi(sidsearch));

        if ((established = strstr(buf, "established")) != NULL) {
            xmlNodePtr estPtr = NULL;
            xmlNodePtr outPtr = NULL;

            estPtr = xmlNewChild(estroot, NULL, "IN", NULL);
            xmlNewProp(estPtr, MAP_STR_TYPE, MAP_STR_ID);
            xmlNewProp(estPtr, MAP_STR_VAL, (char *)sigbuf);

            outPtr = xmlNewChild(estPtr, NULL, "OUT", NULL);
            xmlNewProp(outPtr, MAP_STR_TYPE, MAP_STR_FLAG);
            xmlNewProp(outPtr, MAP_STR_VAL, MAP_STR_ESTABLISHED);
        }

        snort_custom(buf, sigbuf, MAP_STR_SERV);
        snort_custom(buf, sigbuf, MAP_STR_APP);

        // Found no CVE & SID? next rule
        if ((cvesearch = strstr(buf, SIG_CVE)) == NULL &&
            (cvesearch = strstr(buf, SIG_CAN)) == NULL) {
            continue;
        }

        idsInPtr = search_id(sidroot->children, (char *)sigbuf, MAP_STR_VAL); 
        if (idsInPtr == NULL) {
            idsInPtr = xmlNewChild(sidroot, NULL, "IN", NULL);
            xmlNewProp(idsInPtr, MAP_STR_TYPE, MAP_STR_ID);
            xmlNewProp(idsInPtr, MAP_STR_VAL, (char *)sigbuf);
            idsdb.idsIDdbsize++;
        }

        while ((cvefound = strstr(cvesearch, SIG_CVE)) != NULL ||
               (cvefound = strstr(cvesearch, SIG_CAN)) != NULL) {
            xmlNodePtr idsOutPtr = NULL;
            char cvefinal[CVE_STR_LENGTH+1];

            // Copy SID out
            strncpy(cvefinal, cvefound, CVE_STR_LENGTH);
            cvefinal[CVE_STR_LENGTH]= 0;

            idsOutPtr = search_id(idsInPtr->children, cvefinal, MAP_STR_ID); 
            if (idsOutPtr == NULL) {
                idsOutPtr = xmlNewChild(idsInPtr, NULL, "OUT", NULL);
                xmlNewProp(idsOutPtr, MAP_STR_TYPE, MAP_STR_ID);
                xmlNewProp(idsOutPtr, MAP_STR_VAL, (char *)cvefinal);
                idsdb.cvedbsize++;
            }

            cvesearch = cvefound + SIG_LEN;
        }                
        if (cvestart < idsdb.cvedbsize - 1) {
            idsdb.multidsIDs++;
        }
    }
    fclose(fhandle);
}

// add custom ID to service mappings
void snort_custom(const char *buf, const char *sigbuf, char *type)
{
    xmlNodePtr itqmap = NULL;
    xmlNodePtr itqroot = xmlDocGetRootElement(idsdb.idstoqid->doc);
    itqmap = itqroot->children;
    while(itqmap != NULL) {
        xmlNodePtr servfound = NULL;
        char *idsprop = NULL;
        char *maptype = NULL;
        
        if ((maptype = xmlGetProp(itqmap, MAP_STR_TYPE)) == NULL) {
            itqmap = itqmap->next;
            continue;
        }

        if (strstr(type, maptype) == NULL) {
            xmlFree(maptype);
            itqmap = itqmap->next;
            continue;
        }

        if ((idsprop = xmlGetProp(itqmap, MAP_STR_VAL)) == NULL) {
            xmlFree(maptype);
            itqmap = itqmap->next;
            continue;
        }

        if (strstr(buf, idsprop) == NULL) {
            xmlFree(maptype);
            xmlFree(idsprop);
            itqmap = itqmap->next;
            continue;
        }
 
        servfound = search_id(itqroot->children, (char *)sigbuf, MAP_STR_VAL);
        if (servfound == NULL) {
            servfound = xmlNewChild(itqroot, NULL, "IN", NULL);
            xmlNewProp(servfound, MAP_STR_TYPE, type);
            xmlNewProp(servfound, MAP_STR_VAL, (char *)sigbuf);
        }
        
        itq_addqid(servfound, itqmap->children, MAP_STR_CUS);
        
        xmlFree(maptype);
        xmlFree(idsprop);
        itqmap = itqmap->next;
    }
}
