/*
 * $Id: quidscor_util.c,v 1.34 2003/10/02 22:58:18 wkoeppl Exp $
 *
 * Copyright (c) 2003, Qualys, Inc.  All rights reserved.
 *
 * See the file LICENSE for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */
#include "quidscor.h"

extern int debug;
extern int daemonmode;
extern FILE *good;
extern FILE *bad;
extern FILE *unverified;
extern char *logdir;
extern char *pidfile;
extern pid_t cpid;
extern int  inRestart;
extern int  lastsig;

/**
 * sighandler takes care of signales sent to QuIDScor
 */
void sighandler(int sig)
{
    if (lastsig && (sig == SIGALRM)) {
        quid_message(SYS, "Caught expected SIGALRM, exiting [%s]",
                     cpid ? "parent" : "child");
        exit(QSCE_NORMAL);
    }
    lastsig = sig;
    if (sig == SIGUSR1) {
        quid_message(SYS, "Caught SIGUSR1, about to terminate gracefully [%s]",
                     cpid ? "parent" : "child");
        // for some reason the SIGUSR1 is not (always?) enough
        // to interrupt the recvfrom
        signal(SIGALRM, sighandler);
        alarm(1); 
        return;
    }
    if (sig == SIGCHLD && inRestart == TRUE) {
        quid_message(SYS, "Received expected SIGCHLD [%s]",
                     cpid ? "parent" : "child");
        return;
    }
    // notify user why QuIDScor was shut down
    quid_message(SYS, "Caught Signal: %d, Shutting down [%s]", sig,
                 cpid ? "parent" : "child");

    exit(QSCE_NORMAL);
}

/**
 * swith_to_daemon creates a QuIDScor-daemon
 * returns -1 on error / FALSE for parent process / TRUE if daemon is 
 * started
 * returns also TRUE, if daemonmode is deactivated
 */
int switch_to_daemon()
{
    int descr;

    //if daemonmode is deactivated check for pid and return TRUE
    if (!daemonmode) {
        checkpid();
        return TRUE;
    }

    switch (fork()){
    case FORK_SUCCESS:
        break;
    case FORK_ERR:
        return FORK_ERR;
    default:
        return FALSE;
    }
    
    setsid();

    //second fork to prevent zombie-processes (useful? -- dl)
    switch (fork()) {
        case FORK_SUCCESS:
            break;
        case FORK_ERR:
            return FORK_ERR;
        default:
            return FALSE;
    }

    //change to logdir as working directory and check for PID
    chdir(logdir);
    checkpid();

    //close standard descriptors
    for (descr = 0; descr < NUMOF_STAND_DESCR; descr ++) {
        close(descr);
    }

    //open standard descriptors to /dev/null
    descr = open("/dev/null", O_RDWR);
    dup(descr);
    dup(descr);
    umask(027);

    return TRUE;
}

/**
 * checkpid checks, if another QuIDScor is running or not by looking
 * into the specified log-directory and searching for quidscor.pid
 */
void checkpid()
{
    int pidfp;
    int pidid;
    char pidstr[SHO_ARR_LEN];
    char pidbuf[STR_BUF_LEN];

    snprintf(pidbuf, STR_BUF_LEN, "%s/%s", logdir, PID_FILE);
    pidfile = strdup(pidbuf);

    quid_message(DBG, "[D] checking for PID-file: %s", pidfile);

    //try to create pidfile exclusively
    pidfp = open(pidfile, O_CREAT|O_EXCL|O_RDWR, MOD_PID);

    //if everything is ok, write actual PID into file
    if (pidfp >= 0) {
        snprintf(pidstr, SHO_ARR_LEN, "%d\n", getpid());
        write(pidfp, pidstr, strlen(pidstr));
        close(pidfp);
        return;
    }
    close(pidfp);

    //if creation has failed, as write access was checked at first
    //try to access log-directory, read pid from file and check
    //if process exists
    pidfp = open(pidfile, O_RDWR, MOD_PID);
    if (pidfp < 0) {
        quid_message(ERR, "Access to %s denied. Make sure QuIDScor has write and read access.",pidfile);
        quid_exit(QSCE_PID);
    }

    read(pidfp, pidstr, MAX_PIDLEN);
    pidid = atoi(pidstr);

    //if QuIDScor is already running with the PID saved in the file
    //notify user of the PID 
    // TODO: hmmm - should we really kill there ?? (if we do we might
    //       as well unlink the pid file...) - plus : creates: application
    //       bug: quidscor(3979) has SIGCHLD set to SIG_IGN but calls wait().
    if (pidstr && pidid != 0 && kill(pidid,0) == 0) {
        quid_message(ERR, "QuIDScor (PID: %d) is already running.", pidid);
        quid_exit(QSCE_PID);
    }

    //print this if QuIDScor is not running with the PID in the file
    quid_message(ERR, "There is already a pid-file at %s. Please check if QuIDScor was terminated correctly.", pidfile);
    quid_exit(QSCE_PID);
}

void quid_exit(QC_exit code) {
    quid_message(DBG, "Exiting with status %d", code);
//    cleanup();
    exit(code);
}

/**
 * cleanup is called on normal program termination
 */
void cleanup()
{
//    printf("\n---DBG--- : clean up called\n");
    // Parent controlling process isn't actually
    // reading from the socket or writing logs so
    // only the child can unlik the lock file:
    if (cpid <= 0) {
        unlink(pidfile);
    } else if (lastsig != SIGCHLD) {
        // (only if we aren't exiting because the child did already)
        quid_message(SYS, "Requesting child process %d to end", cpid);
        kill(cpid, SIGTERM);
    }
}

/**
 * create_dir creates recursively the specified directory
 * returns NULL on success and name of directory where failure happened
 */
char *create_dir(char *directory)
{
    char *dircheck = NULL;
    char *tok = NULL;
    char *dirbuf = strdup(directory);
    if ((dircheck = calloc(1, sizeof(dirbuf))) == NULL) {
        perror("calloc error at directory creation");
        quid_exit(QSCE_MEMORY);
    }

    //try to create directory starting from root-level
    for (tok = strtok(dirbuf, DIRDELIM);
         (tok != NULL);
         tok = strtok(NULL, DIRDELIM)) {
        // skip empty tokens (multiple spaces, end of line,...)
        dircheck = qg_strappend(dircheck, DIRDELIM);
        dircheck = qg_strappend(dircheck, tok);
        // try to create directory
        if (mkdir(dircheck, CREATE_MOD) != 0 && errno != EEXIST) {
            //if the creation was not successful and the directory does
            //not exist, return name of directory
            return dircheck;
        }
    }

    //free memory allocated at the beginning of the function
    free(dirbuf);
    free(dircheck);
    return NULL;
}

/**
 * reform_path removes '/' from the end of a path -- useful for output
 */
void reform_path(char *path)
{
    int pathlen = 0;
    char *dirbuf = NULL;

    if (!path) {
        return;
    }

    pathlen = strlen(path)-1;
    //the removal is only useful, if path is not an empty string
    if (pathlen >= 0) {
        dirbuf = path+pathlen;
        //replace last '/' with terminating '\0'
        if (*dirbuf == '/') {
            *dirbuf = '\0';
        }
    }
}

/**
 * pmessage creates output if daemon-mode is not enabled
 */
void quid_message(QG_mtype status, const char *message,...)
{
    char temp_buf[STR_BUF_LEN];
    char template[STR_BUF_LEN];
    int print = !daemonmode;
    va_list ap;

    // TODO: FIX ! should only create if we are actually going to
    // do some output (DBG for instance isn't always output/logged)

    // create output-message
    va_start(ap, message);
    vsnprintf(temp_buf, STR_BUF_LEN, message, ap);

    switch(status) {
        // Debug message
        case DBG:
            if (!debug) {
                return;
            }
            // redirect output to syslog, if running in daemonmode
            if (daemonmode) {
                syslog(LOG_NOTICE, temp_buf);
                return;
            }
            snprintf(template, STR_BUF_LEN, "%s\n", temp_buf);
            break;

        // Error message
        case ERR:
            syslog(LOG_NOTICE, temp_buf);
            snprintf(template, STR_BUF_LEN, "[E] %s\n", temp_buf);
            break;

        // SYS level message (info, for syslog)
        case SYS:
            syslog(LOG_NOTICE, temp_buf);
            snprintf(template, STR_BUF_LEN, "[L] %s\n", temp_buf);
            break;
        default:
            return;
    }

    if (print) {
        printf(template);
    }
}

